# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import (Dict, List, Optional, AsyncIterable, Awaitable, Union, TypeVar, AsyncGenerator,
                    TYPE_CHECKING, cast)
from collections import defaultdict
import asyncio
import time

from mautrix.types import UserID, RoomID, EventID, TextMessageEventContent, MessageType
from mautrix.client import Client as MxClient
from mautrix.bridge import BaseUser, async_getter_lock
from mautrix.bridge._community import CommunityHelper, CommunityID
from mautrix.util.simple_lock import SimpleLock
from mautrix.util.opt_prometheus import Summary, Gauge, async_time

from maufbapi import AndroidState, AndroidMQTT, AndroidAPI
from maufbapi.mqtt import Disconnect, Connect, MQTTNotLoggedIn, MQTTNotConnected
from maufbapi.http import InvalidAccessToken
from maufbapi.types import graphql, mqtt as mqtt_t

from .config import Config
from .commands import enter_2fa_code
from .db import User as DBUser, UserPortal, UserContact
from . import portal as po, puppet as pu

METRIC_SYNC_THREADS = Summary('bridge_sync_threads', 'calls to sync_threads')
METRIC_RESYNC = Summary('bridge_on_resync', 'calls to on_resync')
METRIC_UNKNOWN_EVENT = Summary('bridge_on_unknown_event', 'calls to on_unknown_event')
METRIC_MEMBERS_ADDED = Summary('bridge_on_members_added', 'calls to on_members_added')
METRIC_MEMBER_REMOVED = Summary('bridge_on_member_removed', 'calls to on_member_removed')
METRIC_TYPING = Summary('bridge_on_typing', 'calls to on_typing')
METRIC_PRESENCE = Summary('bridge_on_presence', 'calls to on_presence')
METRIC_REACTION = Summary('bridge_on_reaction', 'calls to on_reaction')
METRIC_MESSAGE_UNSENT = Summary('bridge_on_unsent', 'calls to on_unsent')
METRIC_MESSAGE_SEEN = Summary('bridge_on_message_seen', 'calls to on_message_seen')
METRIC_TITLE_CHANGE = Summary('bridge_on_title_change', 'calls to on_title_change')
METRIC_AVATAR_CHANGE = Summary('bridge_on_avatar_change', 'calls to on_avatar_change')
METRIC_THREAD_CHANGE = Summary('bridge_on_thread_change', 'calls to on_thread_change')
METRIC_MESSAGE = Summary('bridge_on_message', 'calls to on_message')
METRIC_LOGGED_IN = Gauge('bridge_logged_in', 'Users logged into the bridge')
METRIC_CONNECTED = Gauge('bridge_connected', 'Bridge users connected to Facebook')

if TYPE_CHECKING:
    from .__main__ import MessengerBridge

try:
    from aiohttp_socks import ProxyError, ProxyConnectionError, ProxyTimeoutError
except ImportError:
    class ProxyError(Exception):
        pass


    ProxyConnectionError = ProxyTimeoutError = ProxyError

T = TypeVar('T')


class User(DBUser, BaseUser):
    temp_disconnect_notices: bool = True
    shutdown: bool = False
    config: Config

    by_mxid: Dict[UserID, 'User'] = {}
    by_fbid: Dict[int, 'User'] = {}

    client: Optional[AndroidAPI]
    mqtt: Optional[AndroidMQTT]
    listen_task: Optional[asyncio.Task]
    seq_id: Optional[int]

    _notice_room_lock: asyncio.Lock
    _notice_send_lock: asyncio.Lock
    is_admin: bool
    permission_level: str
    _is_logged_in: Optional[bool]
    _is_connected: Optional[bool]
    _connection_time: float
    _prev_thread_sync: float
    _prev_reconnect_fail_refresh: float
    _db_instance: Optional[DBUser]
    _sync_lock: SimpleLock
    _is_refreshing: bool

    _community_helper: CommunityHelper
    _community_id: Optional[CommunityID]

    def __init__(self, mxid: UserID, fbid: Optional[int] = None,
                 state: Optional[AndroidState] = None,
                 notice_room: Optional[RoomID] = None) -> None:
        super().__init__(mxid=mxid, fbid=fbid, state=state, notice_room=notice_room)
        self.notice_room = notice_room
        self._notice_room_lock = asyncio.Lock()
        self._notice_send_lock = asyncio.Lock()
        self.command_status = None
        (self.is_whitelisted, self.is_admin,
         self.permission_level) = self.config.get_permissions(mxid)
        self._is_logged_in = None
        self._is_connected = None
        self._connection_time = time.monotonic()
        self._prev_thread_sync = -10
        self._prev_reconnect_fail_refresh = time.monotonic()
        self._community_id = None
        self._sync_lock = SimpleLock("Waiting for thread sync to finish before handling %s",
                                     log=self.log)
        self._is_refreshing = False
        self._metric_value = defaultdict(lambda: False)
        self.dm_update_lock = asyncio.Lock()

        self.log = self.log.getChild(self.mxid)

        self.client = None
        self.mqtt = None
        self.listen_task = None
        self.seq_id = None

    @classmethod
    def init_cls(cls, bridge: 'MessengerBridge') -> AsyncIterable[Awaitable[bool]]:
        cls.bridge = bridge
        cls.config = bridge.config
        cls.az = bridge.az
        cls.loop = bridge.loop
        cls._community_helper = CommunityHelper(cls.az)
        cls.temp_disconnect_notices = bridge.config["bridge.temporary_disconnect_notices"]
        return (user.load_session() async for user in cls.all_logged_in())

    @property
    def is_connected(self) -> Optional[bool]:
        return self._is_connected

    @is_connected.setter
    def is_connected(self, val: Optional[bool]) -> None:
        if self._is_connected != val:
            self._is_connected = val
            self._connection_time = time.monotonic()

    # region Database getters

    def _add_to_cache(self) -> None:
        self.by_mxid[self.mxid] = self
        if self.fbid:
            self.by_fbid[self.fbid] = self

    @classmethod
    async def all_logged_in(cls) -> AsyncGenerator['User', None]:
        users = await super().all_logged_in()
        user: cls
        for user in users:
            try:
                yield cls.by_mxid[user.mxid]
            except KeyError:
                user._add_to_cache()
                yield user

    @classmethod
    @async_getter_lock
    async def get_by_mxid(cls, mxid: UserID, *, create: bool = True) -> Optional['User']:
        if pu.Puppet.get_id_from_mxid(mxid) or mxid == cls.az.bot_mxid:
            return None
        try:
            return cls.by_mxid[mxid]
        except KeyError:
            pass

        user = cast(cls, await super().get_by_mxid(mxid))
        if user is not None:
            user._add_to_cache()
            return user

        if create:
            cls.log.debug(f"Creating user instance for {mxid}")
            user = cls(mxid)
            await user.insert()
            user._add_to_cache()
            return user

        return None

    @classmethod
    @async_getter_lock
    async def get_by_fbid(cls, fbid: int) -> Optional['User']:
        try:
            return cls.by_fbid[fbid]
        except KeyError:
            pass

        user = cast(cls, await super().get_by_fbid(fbid))
        if user is not None:
            user._add_to_cache()
            return user

        return None

    # endregion

    async def load_session(self, _override: bool = False, _raise_errors: bool = False) -> bool:
        if self._is_logged_in and not _override:
            return True
        elif not self.state:
            return False
        attempt = 0
        client = AndroidAPI(self.state, log=self.log.getChild("api"))
        while True:
            try:
                user_info = await client.get_self()
                break
            except (ProxyError, ProxyTimeoutError, ProxyConnectionError, ConnectionError) as e:
                attempt += 1
                wait = min(attempt * 10, 60)
                self.log.warning(f"{e.__class__.__name__} while trying to restore session, "
                                 f"retrying in {wait} seconds: {e}")
                await asyncio.sleep(wait)
            except Exception:
                self.log.exception("Failed to restore session")
                if _raise_errors:
                    raise
                return False
        if user_info:
            self.log.info("Loaded session successfully")
            self.client = client
            self._track_metric(METRIC_LOGGED_IN, True)
            self._is_logged_in = True
            self.is_connected = None
            self.stop_listen()
            asyncio.ensure_future(self.post_login(), loop=self.loop)
            return True
        return False

    async def is_logged_in(self, _override: bool = False) -> bool:
        if not self.state or not self.client:
            return False
        if self._is_logged_in is None or _override:
            try:
                self._is_logged_in = bool(await self.client.get_self())
            except Exception:
                self.log.exception("Exception checking login status")
                self._is_logged_in = False
        return self._is_logged_in

    async def try_refresh(self) -> None:
        try:
            await self.refresh()
        except InvalidAccessToken as e:
            self.log.exception("Invalid auth error while trying to refresh after connection error")
            await self.send_bridge_notice("Got authentication error from Messenger:\n\n"
                                          f"> {e!s}\n\n"
                                          "If you changed your Facebook password or enabled two-"
                                          "factor authentication, this is normal and you just "
                                          "need to log in again.",
                                          important=True)
            await self.logout(remove_fbid=False)
        except Exception:
            self.log.exception("Fatal error while trying to refresh after connection error")
            await self.send_bridge_notice("Fatal error while trying to refresh after connection "
                                          "error (see logs for more info)", important=True)

    async def refresh(self, force_notice: bool = False) -> None:
        event_id = None
        self._is_refreshing = True
        if self.mqtt:
            self.log.debug("Disconnecting MQTT connection for session refresh...")
            if self.temp_disconnect_notices or force_notice:
                event_id = await self.send_bridge_notice("Disconnecting Messenger MQTT connection "
                                                         "for session refresh...")
            self.mqtt.disconnect()
            if self.listen_task:
                try:
                    await asyncio.wait_for(self.listen_task, timeout=3)
                except asyncio.TimeoutError:
                    self.log.debug("Waiting for MQTT connection timed out")
                else:
                    self.log.debug("MQTT connection disconnected")
            self.mqtt = None
            if self.client:
                self.client.sequence_id_callback = None
        if self.temp_disconnect_notices or force_notice:
            event_id = await self.send_bridge_notice("Refreshing session...", edit=event_id)
        try:
            await self.load_session(_override=True, _raise_errors=True)
        except Exception:
            await self.send_bridge_notice("Failed to refresh Messenger session: unknown error "
                                          "(see logs for more details)", edit=event_id)
        finally:
            self._is_refreshing = False

    async def reconnect(self) -> None:
        self._is_refreshing = True
        self.mqtt.disconnect()
        await self.listen_task
        self.listen_task = None
        self.mqtt = None
        self.start_listen()
        self._is_refreshing = False

    async def logout(self, remove_fbid: bool = True) -> bool:
        ok = True
        self.stop_listen()
        if self.state:
            # TODO is there even a logout API for messenger mobile?
            pass
            # try:
            #     await self.session.logout()
            # except fbchat.FacebookError:
            #     self.log.exception("Error while logging out")
            #     ok = False
        self._track_metric(METRIC_LOGGED_IN, False)
        self.state = None
        self._is_logged_in = False
        self.is_connected = None
        self.client = None
        self.mqtt = None

        if self.fbid and remove_fbid:
            await UserContact.delete_all(self.fbid)
            await UserPortal.delete_all(self.fbid)
            del self.by_fbid[self.fbid]
            self.fbid = None

        await self.save()
        return ok

    async def post_login(self) -> None:
        self.log.info("Running post-login actions")
        self._add_to_cache()

        try:
            puppet = await pu.Puppet.get_by_fbid(self.fbid)

            if puppet.custom_mxid != self.mxid and puppet.can_auto_login(self.mxid):
                self.log.info(f"Automatically enabling custom puppet")
                await puppet.switch_mxid(access_token="auto", mxid=self.mxid)
        except Exception:
            self.log.exception("Failed to automatically enable custom puppet")

        await self._create_community()
        await self.sync_threads()
        self.start_listen()

    async def _create_community(self) -> None:
        template = self.config["bridge.community_template"]
        if not template:
            return
        localpart, server = MxClient.parse_user_id(self.mxid)
        community_localpart = template.format(localpart=localpart, server=server)
        self.log.debug(f"Creating personal filtering community {community_localpart}...")
        self._community_id, created = await self._community_helper.create(community_localpart)
        if created:
            await self._community_helper.update(self._community_id, name="Facebook Messenger",
                                                avatar_url=self.config["appservice.bot_avatar"],
                                                short_desc="Your Facebook bridged chats")
            await self._community_helper.invite(self._community_id, self.mxid)

    async def _add_community(self, up: Optional[UserPortal], contact: Optional[UserContact],
                             portal: 'po.Portal', puppet: Optional['pu.Puppet']) -> None:
        if portal.mxid:
            if not up or not up.in_community:
                ic = await self._community_helper.add_room(self._community_id, portal.mxid)
                if up and ic:
                    up.in_community = True
                    await up.save()
                elif not up:
                    await UserPortal(user=self.fbid, in_community=ic, portal=portal.fbid,
                                     portal_receiver=portal.fb_receiver).insert()
        if puppet:
            await self._add_community_puppet(contact, puppet)

    async def _add_community_puppet(self, contact: Optional[UserContact],
                                    puppet: 'pu.Puppet') -> None:
        if not contact or not contact.in_community:
            await puppet.default_mxid_intent.ensure_registered()
            ic = await self._community_helper.join(self._community_id,
                                                   puppet.default_mxid_intent)
            if contact and ic:
                contact.in_community = True
                await contact.save()
            elif not contact:
                # This uses upsert instead of insert as a hacky fix for potential conflicts
                await UserContact(user=self.fbid, contact=puppet.fbid, in_community=ic).upsert()

    async def get_direct_chats(self) -> Dict[UserID, List[RoomID]]:
        return {
            pu.Puppet.get_mxid_from_id(portal.fbid): [portal.mxid]
            async for portal in po.Portal.get_all_by_receiver(self.fbid)
            if portal.mxid
        }

    @async_time(METRIC_SYNC_THREADS)
    async def sync_threads(self) -> None:
        if self._prev_thread_sync + 10 > time.monotonic() and self.mqtt.seq_id is not None:
            self.log.debug("Previous thread sync was less than 10 seconds ago, not re-syncing")
            return
        self._prev_thread_sync = time.monotonic()
        try:
            await self._sync_threads()
        except Exception:
            self.log.exception("Failed to sync threads")

    async def _sync_threads(self) -> None:
        sync_count = self.config["bridge.initial_chat_sync"]
        self.log.debug("Fetching threads...")
        ups = await UserPortal.all(self.fbid)
        contacts = await UserContact.all(self.fbid)
        # TODO paginate with 20 threads per request
        resp = await self.client.fetch_thread_list(thread_count=sync_count)
        self.seq_id = int(resp.sync_sequence_id)
        if self.mqtt:
            self.mqtt.seq_id = self.seq_id
        if sync_count <= 0:
            return
        for thread in resp.nodes:
            try:
                await self._sync_thread(thread, ups, contacts)
            except Exception:
                self.log.exception("Failed to sync thread %s", thread.id)

        await self.update_direct_chats()

    async def _sync_thread(self, thread: graphql.Thread, ups: Dict[int, UserPortal],
                           contacts: Dict[int, UserContact]) -> None:
        self.log.debug(f"Syncing thread {thread.thread_key.id}")
        is_direct = bool(thread.thread_key.other_user_id)
        portal = await po.Portal.get_by_thread(thread.thread_key, self.fbid)
        puppet = (await pu.Puppet.get_by_fbid(thread.thread_key.other_user_id)
                  if is_direct else None)

        await self._add_community(ups.get(portal.fbid, None),
                                  contacts.get(puppet.fbid, None) if puppet else None,
                                  portal, puppet)

        if not portal.mxid:
            await portal.create_matrix_room(self, thread)
        else:
            await portal.update_matrix_room(self, thread)
            await portal.backfill(self, is_initial=False, thread=thread)

    async def is_in_portal(self, portal: 'po.Portal') -> bool:
        return await UserPortal.get(self.fbid, portal.fbid, portal.fb_receiver) is not None

    async def on_2fa_callback(self) -> str:
        if self.command_status and self.command_status.get("action", "") == "Login":
            future = self.loop.create_future()
            self.command_status["future"] = future
            self.command_status["next"] = enter_2fa_code
            await self.az.intent.send_notice(self.command_status["room_id"],
                                             "You have two-factor authentication enabled. "
                                             "Please send the code here.")
            return await future
        raise RuntimeError("No ongoing login command")

    async def get_notice_room(self) -> RoomID:
        if not self.notice_room:
            async with self._notice_room_lock:
                # If someone already created the room while this call was waiting,
                # don't make a new room
                if self.notice_room:
                    return self.notice_room
                self.notice_room = await self.az.intent.create_room(
                    is_direct=True, invitees=[self.mxid],
                    topic="Facebook Messenger bridge notices")
                await self.save()
        return self.notice_room

    async def send_bridge_notice(self, text: str, edit: Optional[EventID] = None,
                                 important: bool = False) -> Optional[EventID]:
        event_id = None
        try:
            self.log.debug("Sending bridge notice: %s", text)
            content = TextMessageEventContent(body=text, msgtype=(MessageType.TEXT if important
                                                                  else MessageType.NOTICE))
            if edit:
                content.set_edit(edit)
            # This is locked to prevent notices going out in the wrong order
            async with self._notice_send_lock:
                event_id = await self.az.intent.send_message(await self.get_notice_room(), content)
        except Exception:
            self.log.warning("Failed to send bridge notice", exc_info=True)
        return edit or event_id

    # region Facebook event handling

    def start_listen(self) -> None:
        self.listen_task = self.loop.create_task(self._try_listen())

    def _disconnect_listener_after_error(self) -> None:
        try:
            self.mqtt.disconnect()
        except Exception:
            self.log.debug("Error disconnecting listener after error", exc_info=True)

    def _update_seq_id(self, seq_id: int) -> None:
        self.seq_id = seq_id

    def _update_region_hint(self, region_hint: str) -> None:
        self.log.debug(f"Got region hint {region_hint}")
        self.state.session.region_hint = region_hint
        self.loop.create_task(self.save())

    async def _try_listen(self) -> None:
        try:
            if not self.mqtt:
                self.mqtt = AndroidMQTT(self.state, log=self.log.getChild("mqtt"))
                self.mqtt.seq_id_update_callback = self._update_seq_id
                self.mqtt.region_hint_callback = self._update_region_hint
                self.mqtt.add_event_handler(mqtt_t.Message, self.on_message)
                self.mqtt.add_event_handler(mqtt_t.ExtendedMessage, self.on_message)
                self.mqtt.add_event_handler(mqtt_t.NameChange, self.on_title_change)
                self.mqtt.add_event_handler(mqtt_t.AvatarChange, self.on_avatar_change)
                self.mqtt.add_event_handler(mqtt_t.UnsendMessage, self.on_message_unsent)
                self.mqtt.add_event_handler(mqtt_t.ReadReceipt, self.on_message_seen)
                self.mqtt.add_event_handler(mqtt_t.OwnReadReceipt, self.on_message_seen_self)
                self.mqtt.add_event_handler(mqtt_t.Reaction, self.on_reaction)
                self.mqtt.add_event_handler(mqtt_t.AddMember, self.on_members_added)
                self.mqtt.add_event_handler(mqtt_t.RemoveMember, self.on_member_removed)
                self.mqtt.add_event_handler(mqtt_t.ThreadChange, self.on_thread_change)
                self.mqtt.add_event_handler(mqtt_t.MessageSyncError, self.on_message_sync_error)
                self.mqtt.add_event_handler(Connect, self.on_connect)
                self.mqtt.add_event_handler(Disconnect, self.on_disconnect)
            await self.mqtt.listen(self.seq_id)
            self.is_connected = False
            if not self._is_refreshing and not self.shutdown:
                await self.send_bridge_notice("Facebook Messenger connection closed without error")
        except (MQTTNotLoggedIn, MQTTNotConnected) as e:
            self.log.debug("Listen threw a Facebook error", exc_info=True)
            refresh = (self.config["bridge.refresh_on_reconnection_fail"]
                       and self._prev_reconnect_fail_refresh + 120 < time.monotonic())
            next_action = ("Refreshing session..." if refresh else "Not retrying!")
            event = ("Disconnected from" if isinstance(e, MQTTNotLoggedIn)
                     else "Failed to connect to")
            message = f"{event} Facebook Messenger: {e}. {next_action}"
            self.log.warning(message)
            if not refresh or self.temp_disconnect_notices:
                await self.send_bridge_notice(message, important=not refresh)
            if refresh:
                self._prev_reconnect_fail_refresh = time.monotonic()
                self.loop.create_task(self.try_refresh())
            else:
                self._disconnect_listener_after_error()
        except Exception:
            self.is_connected = False
            self.log.exception("Fatal error in listener")
            await self.send_bridge_notice("Fatal error in listener (see logs for more info)",
                                          important=True)
            self._disconnect_listener_after_error()

    # @async_time(METRIC_UNKNOWN_EVENT)
    # async def on_unknown_event(self, evt: fbchat.UnknownEvent) -> None:
    #     self.log.debug(f"Unknown event %s: %s", evt.source, evt.data)

    async def on_connect(self, evt: Connect) -> None:
        now = time.monotonic()
        disconnected_at = self._connection_time
        max_delay = self.config["bridge.resync_max_disconnected_time"]
        first_connect = self.is_connected is None
        self.is_connected = True
        self._track_metric(METRIC_CONNECTED, True)
        if not first_connect and disconnected_at + max_delay < now:
            duration = int(now - disconnected_at)
            self.log.debug("Disconnection lasted %d seconds, not re-syncing threads...", duration)
        elif self.temp_disconnect_notices:
            await self.send_bridge_notice("Connected to Facebook Messenger")

    async def on_disconnect(self, evt: Disconnect) -> None:
        self.is_connected = False
        self._track_metric(METRIC_CONNECTED, False)
        if self.temp_disconnect_notices:
            await self.send_bridge_notice(f"Disconnected from Facebook Messenger: {evt.reason}")

    # @async_time(METRIC_RESYNC)
    # async def on_resync(self, evt: fbchat.Resync) -> None:
    #     self.log.info("sequence_id changed, resyncing threads...")
    #     await self.sync_threads()

    def stop_listen(self) -> None:
        if self.mqtt:
            self.mqtt.disconnect()
        if self.listen_task:
            self.listen_task.cancel()
        self.mqtt = None
        self.listen_task = None

    async def on_logged_in(self, state: AndroidState) -> None:
        self.fbid = state.session.uid
        self.state = state
        self.client = AndroidAPI(state, log=self.log.getChild("api"))
        await self.save()
        self.stop_listen()
        asyncio.ensure_future(self.post_login(), loop=self.loop)

    @async_time(METRIC_MESSAGE)
    async def on_message(self, evt: Union[mqtt_t.Message, mqtt_t.ExtendedMessage]) -> None:
        if isinstance(evt, mqtt_t.ExtendedMessage):
            reply_to = evt.reply_to_message
            evt = evt.message
        else:
            reply_to = None
        portal = await po.Portal.get_by_thread(evt.metadata.thread, self.fbid)
        puppet = await pu.Puppet.get_by_fbid(evt.metadata.sender)
        # if not puppet.name:
        #     await puppet.update_info(self)
        await portal.backfill_lock.wait(evt.metadata.id)
        await portal.handle_facebook_message(self, puppet, evt, reply_to=reply_to)

    @async_time(METRIC_TITLE_CHANGE)
    async def on_title_change(self, evt: mqtt_t.NameChange) -> None:
        portal = await po.Portal.get_by_thread(evt.metadata.thread, self.fbid)
        sender = await pu.Puppet.get_by_fbid(evt.metadata.sender)
        await portal.backfill_lock.wait("title change")
        await portal.handle_facebook_name(self, sender, evt.new_name, evt.metadata.id,
                                          evt.metadata.timestamp)

    @async_time(METRIC_AVATAR_CHANGE)
    async def on_avatar_change(self, evt: mqtt_t.AvatarChange) -> None:
        portal = await po.Portal.get_by_thread(evt.metadata.thread, self.fbid)
        sender = await pu.Puppet.get_by_fbid(evt.metadata.sender)
        await portal.backfill_lock.wait("avatar change")
        await portal.handle_facebook_photo(self, sender, evt.new_avatar, evt.metadata.id,
                                           evt.metadata.timestamp)

    @async_time(METRIC_MESSAGE_SEEN)
    async def on_message_seen(self, evt: mqtt_t.ReadReceipt) -> None:
        puppet = await pu.Puppet.get_by_fbid(evt.user_id)
        portal = await po.Portal.get_by_thread(evt.thread, self.fbid, create=False)
        if portal and portal.mxid:
            await portal.backfill_lock.wait(f"read receipt from {puppet.fbid}")
            await portal.handle_facebook_seen(self, puppet, evt.read_to)

    @async_time(METRIC_MESSAGE_SEEN)
    async def on_message_seen_self(self, evt: mqtt_t.OwnReadReceipt) -> None:
        puppet = await pu.Puppet.get_by_fbid(self.fbid)
        for thread in evt.threads:
            portal = await po.Portal.get_by_thread(thread, self.fbid, create=False)
            if portal:
                await portal.backfill_lock.wait(f"read receipt from {puppet.fbid}")
                await portal.handle_facebook_seen(self, puppet, evt.read_to)

    @async_time(METRIC_MESSAGE_UNSENT)
    async def on_message_unsent(self, evt: mqtt_t.UnsendMessage) -> None:
        portal = await po.Portal.get_by_thread(evt.thread, self.fbid, create=False)
        if portal and portal.mxid:
            await portal.backfill_lock.wait(f"redaction of {evt.message_id}")
            puppet = await pu.Puppet.get_by_fbid(evt.user_id)
            await portal.handle_facebook_unsend(puppet, evt.message_id, timestamp=evt.timestamp)

    @async_time(METRIC_REACTION)
    async def on_reaction(self, evt: mqtt_t.Reaction) -> None:
        portal = await po.Portal.get_by_thread(evt.thread, self.fbid, create=False)
        if not portal or not portal.mxid:
            return
        puppet = await pu.Puppet.get_by_fbid(evt.reaction_sender_id)
        await portal.backfill_lock.wait(f"reaction to {evt.message_id}")
        if evt.reaction is None:
            await portal.handle_facebook_reaction_remove(self, puppet, evt.message_id)
        else:
            await portal.handle_facebook_reaction_add(self, puppet, evt.message_id, evt.reaction)

    # @async_time(METRIC_PRESENCE)
    # async def on_presence(self, evt: fbchat.Presence) -> None:
    #     for user, status in evt.statuses.items():
    #         puppet = pu.Puppet.get_by_fbid(user, create=False)
    #         if puppet:
    #             await puppet.default_mxid_intent.set_presence(
    #                 presence=PresenceState.ONLINE if status.active else PresenceState.OFFLINE,
    #                 ignore_cache=True)
    #
    # @async_time(METRIC_TYPING)
    # async def on_typing(self, evt: fbchat.Typing) -> None:
    #     fb_receiver = self.fbid if isinstance(evt.thread, fbchat.User) else None
    #     portal = po.Portal.get_by_thread(evt.thread, fb_receiver)
    #     if portal.mxid and not portal.backfill_lock.locked:
    #         puppet = pu.Puppet.get_by_fbid(evt.author.id)
    #         await puppet.intent.set_typing(portal.mxid, is_typing=evt.status, timeout=120000)

    @async_time(METRIC_MEMBERS_ADDED)
    async def on_members_added(self, evt: mqtt_t.AddMember) -> None:
        portal = await po.Portal.get_by_thread(evt.metadata.thread, self.fbid)
        if portal.mxid:
            sender = await pu.Puppet.get_by_fbid(evt.metadata.sender)
            users = [await pu.Puppet.get_by_fbid(user.id) for user in evt.users]
            await portal.backfill_lock.wait("member add")
            await portal.handle_facebook_join(self, sender, users)

    @async_time(METRIC_MEMBER_REMOVED)
    async def on_member_removed(self, evt: mqtt_t.RemoveMember) -> None:
        portal = await po.Portal.get_by_thread(evt.metadata.thread, self.fbid)
        if portal.mxid:
            sender = await pu.Puppet.get_by_fbid(evt.metadata.sender)
            user = await pu.Puppet.get_by_fbid(evt.user_id)
            await portal.backfill_lock.wait("member remove")
            await portal.handle_facebook_leave(self, sender, user)

    @async_time(METRIC_THREAD_CHANGE)
    async def on_thread_change(self, evt: mqtt_t.ThreadChange) -> None:
        pass
        # portal = await po.Portal.get_by_thread(evt.metadata.thread, self.fbid)
        # if not portal.mxid:
        #     return

        # TODO
        # if evt.action == mqtt_t.ThreadChangeAction.ADMINS:
        #     sender = await pu.Puppet.get_by_fbid(evt.metadata.sender)
        #     user = await pu.Puppet.get_by_fbid(evt.action_data["TARGET_ID"])
        #     make_admin = evt.action_data["ADMIN_EVENT"] == "add_admin"
        #     # TODO does the ADMIN_TYPE data matter?
        #     await portal.backfill_lock.wait("admin change")
        #     await portal.handle_facebook_admin(self, sender, user, make_admin)

    async def on_message_sync_error(self, evt: mqtt_t.MessageSyncError) -> None:
        self.log.error(f"Message sync error: {evt.value}, resyncing...")
        await self.send_bridge_notice(f"Message sync error: {evt.value}, resyncing...")
        self.stop_listen()
        await self.sync_threads()
        self.start_listen()

    # endregion



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, List, TYPE_CHECKING, ClassVar

from asyncpg import Record
from attr import dataclass

from mautrix.types import RoomID, EventID
from mautrix.util.async_db import Database

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class Message:
    db: ClassVar[Database] = fake_db

    mxid: EventID
    mx_room: RoomID
    fbid: str
    fb_chat: int
    fb_receiver: int
    index: int
    timestamp: int

    @classmethod
    def _from_row(cls, row: Optional[Record]) -> Optional['Message']:
        if row is None:
            return None
        return cls(**row)

    @classmethod
    async def get_all_by_fbid(cls, fbid: str, fb_receiver: int) -> List['Message']:
        q = ("SELECT mxid, mx_room, fbid, fb_chat, fb_receiver, index, timestamp "
             "FROM message WHERE fbid=$1 AND fb_receiver=$2")
        rows = await cls.db.fetch(q, fbid, fb_receiver)
        return [cls._from_row(row) for row in rows]

    @classmethod
    async def get_by_fbid(cls, fbid: str, fb_receiver: int, index: int = 0) -> Optional['Message']:
        q = ("SELECT mxid, mx_room, fbid, fb_chat, fb_receiver, index, timestamp "
             "FROM message WHERE fbid=$1 AND fb_receiver=$2 AND index=$3")
        row = await cls.db.fetchrow(q, fbid, fb_receiver, index)
        return cls._from_row(row)

    @classmethod
    async def delete_all_by_room(cls, room_id: RoomID) -> None:
        await cls.db.execute("DELETE FROM message WHERE mx_room=$1", room_id)

    @classmethod
    async def get_by_mxid(cls, mxid: EventID, mx_room: RoomID) -> Optional['Message']:
        q = ("SELECT mxid, mx_room, fbid, fb_chat, fb_receiver, index, timestamp "
             "FROM message WHERE mxid=$1 AND mx_room=$2")
        row = await cls.db.fetchrow(q, mxid, mx_room)
        return cls._from_row(row)

    @classmethod
    async def get_most_recent(cls, fb_chat: int, fb_receiver: int) -> Optional['Message']:
        q = ("SELECT mxid, mx_room, fbid, fb_chat, fb_receiver, index, timestamp "
             "FROM message WHERE fb_chat=$1 AND fb_receiver=$2 ORDER BY timestamp DESC LIMIT 1")
        row = await cls.db.fetchrow(q, fb_chat, fb_receiver)
        return cls._from_row(row)

    @classmethod
    async def get_closest_before(cls, fb_chat: int, fb_receiver: int, timestamp: int
                                 ) -> Optional['Message']:
        q = ("SELECT mxid, mx_room, fbid, fb_chat, fb_receiver, index, timestamp "
             "FROM message WHERE fb_chat=$1 AND fb_receiver=$2 AND timestamp<=$3 "
             "ORDER BY timestamp DESC LIMIT 1")
        row = await cls.db.fetchrow(q, fb_chat, fb_receiver, timestamp)
        return cls._from_row(row)

    @classmethod
    async def bulk_create(cls, fbid: str, fb_chat: int, fb_receiver: int, event_ids: List[EventID],
                          timestamp: int, mx_room: RoomID) -> None:
        if not event_ids:
            return
        columns = ["mxid", "mx_room", "fbid", "fb_chat", "fb_receiver", "index", "timestamp"]
        records = [(mxid,   mx_room,   fbid,   fb_chat,   fb_receiver,   index,   timestamp)
                   for index, mxid in enumerate(event_ids)]
        async with cls.db.acquire() as conn, conn.transaction():
            await conn.copy_records_to_table("message", records=records, columns=columns)

    async def insert(self) -> None:
        q = ("INSERT INTO message (mxid, mx_room, fbid, fb_chat, fb_receiver, index, timestamp) "
             "VALUES ($1, $2, $3, $4, $5, $6, $7)")
        await self.db.execute(q, self.mxid, self.mx_room, self.fbid, self.fb_chat,
                              self.fb_receiver, self.index, self.timestamp)

    async def delete(self) -> None:
        q = "DELETE FROM message WHERE fbid=$1 AND fb_receiver=$2 AND index=$3"
        await self.db.execute(q, self.fbid, self.fb_receiver, self.index)

    async def delete_all(self) -> None:
        q = "DELETE FROM message WHERE fbid=$1 AND fb_receiver=$2"
        await self.db.execute(q, self.fbid, self.fb_receiver)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Optional, Union, Dict
from uuid import uuid4

from yarl import URL

from mautrix.types import JSON

from ..types import (ThreadListResponse, ThreadListQuery, MessageList, MoreMessagesQuery,
                     FetchStickersWithPreviewsQuery, StickerPreviewResponse, MessageUndoSend,
                     MessageUnsendResponse, ReactionAction, MessageReactionMutation,
                     DownloadImageFragment, ImageFragment, SubsequentMediaResponse,
                     SubsequentMediaQuery, FbIdToCursorQuery, FileAttachmentUrlQuery,
                     FileAttachmentURLResponse, SearchEntitiesNamedQuery, SearchEntitiesResponse,
                     ThreadQuery, ThreadQueryResponse)
from ..types.graphql import PageInfo, ThreadMessageID, OwnInfo, Thread
from .base import BaseAndroidAPI
from .login import LoginAPI
from .upload import UploadAPI


class AndroidAPI(LoginAPI, UploadAPI, BaseAndroidAPI):
    _file_url_cache: Dict[ThreadMessageID, FileAttachmentURLResponse]

    async def fetch_thread_list(self, **kwargs) -> ThreadListResponse:
        return await self.graphql(ThreadListQuery(**kwargs), response_type=ThreadListResponse,
                                  path=["data", "viewer", "message_threads"])

    async def fetch_thread_info(self, *thread_ids: Union[str, int], **kwargs) -> List[Thread]:
        resp = await self.graphql(ThreadQuery(thread_ids=[str(i) for i in thread_ids], **kwargs),
                                  path=["data"], response_type=ThreadQueryResponse)
        return resp.message_threads

    async def fetch_messages(self, thread_id: int, before_time_ms: int, **kwargs
                             ) -> MessageList:
        return await self.graphql(MoreMessagesQuery(thread_id=str(thread_id),
                                                    before_time_ms=str(before_time_ms), **kwargs),
                                  path=["data", "message_thread", "messages"],
                                  response_type=MessageList)

    async def fetch_stickers(self, ids: List[int], **kwargs) -> StickerPreviewResponse:
        kwargs["sticker_ids"] = [str(id) for id in ids]
        return await self.graphql(FetchStickersWithPreviewsQuery(**kwargs),
                                  path=["data"], response_type=StickerPreviewResponse, b=True)

    async def unsend(self, message_id: str) -> MessageUnsendResponse:
        return await self.graphql(MessageUndoSend(message_id=message_id,
                                                  client_mutation_id=str(uuid4()),
                                                  actor_id=str(self.state.session.uid)),
                                  path=["data", "message_undo_send"],
                                  response_type=MessageUnsendResponse)

    async def react(self, message_id: str, reaction: Optional[str]) -> None:
        action = ReactionAction.ADD if reaction else ReactionAction.REMOVE
        await self.graphql(MessageReactionMutation(message_id=message_id, reaction=reaction,
                                                   action=action, client_mutation_id=str(uuid4()),
                                                   actor_id=str(self.state.session.uid)),
                           response_type=JSON)

    async def fetch_image(self, media_id: Union[int, str]) -> ImageFragment:
        return await self.graphql(DownloadImageFragment(fbid=str(media_id)),
                                  path=["data", "node"], response_type=ImageFragment)

    async def fbid_to_cursor(self, thread_id: Union[int, str], media_id: Union[int, str]
                             ) -> PageInfo:
        return await self.graphql(FbIdToCursorQuery(fbid=str(media_id), thread_id=str(thread_id)),
                                  path=["data", "message_thread", "message_shared_media",
                                        "page_info"], response_type=PageInfo)

    async def media_query(self, thread_id: Union[int, str], cursor: Optional[str] = None
                          ) -> SubsequentMediaResponse:
        return await self.graphql(SubsequentMediaQuery(thread_id=str(thread_id), cursor_id=cursor),
                                  path=["data", "message_thread", "mediaResult"],
                                  response_type=SubsequentMediaResponse)

    async def search(self, query: str, **kwargs) -> SearchEntitiesResponse:
        return await self.graphql(SearchEntitiesNamedQuery(search_query=query, **kwargs),
                                  path=["data", "entities_named"],
                                  response_type=SearchEntitiesResponse)

    async def get_image_url(self, message_id: str, attachment_id: Union[int, str],
                            preview: bool = False, max_width: int = 384,
                            max_height: int = 480) -> Optional[str]:
        query = {
            "method": "POST",
            "redirect": "true",
            "access_token": self.state.session.access_token,
            "mid": f"m_{message_id}",
            "aid": str(attachment_id),
        }
        if preview:
            query["preview"] = "1"
            query["max_width"] = max_width
            query["max_height"] = max_height
        headers = {
            "referer": f"fbapp://{self.state.application.client_id}/messenger_thread_photo",
            "x-fb-friendly-name": "image",
        }
        resp = await self.get((self.graph_url / "messaging_get_attachment").with_query(query),
                              headers=headers, include_auth=False, allow_redirects=False)
        # TODO handle errors more properly?
        try:
            return resp.headers["Location"]
        except KeyError:
            return None

    async def get_file_url(self, thread_id: Union[str, int], message_id: str,
                           attachment_id: Union[str, int]) -> Optional[URL]:
        attachment_id = str(attachment_id)
        msg_id = ThreadMessageID(thread_id=str(thread_id), message_id=message_id)
        try:
            resp = self._file_url_cache[msg_id]
        except KeyError:
            resp = await self.graphql(FileAttachmentUrlQuery(thread_msg_id=msg_id),
                                      path=["data", "message"],
                                      response_type=FileAttachmentURLResponse)
            if len(resp.blob_attachments) > 1:
                self._file_url_cache[msg_id] = resp
        for attachment in resp.blob_attachments:
            if attachment.attachment_fbid == attachment_id:
                url = URL(resp.blob_attachments[0].url)
                if url.host == "l.facebook.com":
                    url = URL(url.query["u"])
                return url
        return None

    async def get_self(self) -> OwnInfo:
        async with self.get(self.graph_url / str(self.state.session.uid)) as resp:
            json_data = await self._handle_response(resp)
        return OwnInfo.deserialize(json_data)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List

from attr import dataclass

from maufbapi.thrift import TType, ThriftObject, field, autospec


@autospec
@dataclass(kw_only=True)
class SendMessageRequest(ThriftObject):
    # tfbid_<groupid> for groups, plain user id for users
    chat_id: str
    message: str
    offline_threading_id: int = field(TType.I64)
    # index 4: ???
    # Example values:
    #   'is_in_chatheads': 'false'
    #   'ld': '{"u":1674434.........}'
    #   'entrypoint': 'messenger_inbox:in_thread'
    #   'trigger': '2:thread_list:thread' or 'thread_view_messages_fragment_unknown'
    #   'active_now': '{"is_online":"false","last_active_seconds":"1431"}'
    #
    #   'media_camera_mode': 'VIDEO'
    #   'trigger': 'thread_view_messages_fragment_unknown'
    #   'entry_point': 'THREAD_CAMERA_COMPOSER_BUTTON'
    flags: Dict[str, str] = field(index=5, factory=lambda: {})
    # 369239263222822 = "like"
    sticker: str = field(default=None)
    # indices 7 and 8: ???
    media_ids: List[str] = field(default=None)
    # indices 10 and 11: ???
    sender_id: int = field(TType.I64, index=12)
    # indices 13-17: ???
    unknown_int32: int = field(TType.I32, index=18, default=0)
    # index 19: ???
    extra_metadata: Dict[str, str] = field(index=20, default=None)
    unknown_int64: int = field(TType.I64, index=21, default=0)
    # index 22: ???
    unknown_bool: bool = field(TType.BOOL, index=23, default=False)
    # this is weird int64 that looks like offline_threading_id, but isn't quite the same
    tid2: int = field(TType.I64, index=24)
    # indices 25-27: ???
    reply_to: str = field(index=28)


@autospec
@dataclass(kw_only=True)
class MarkReadRequest(ThriftObject):
    receipt_type: str = "read"
    unknown_boolean: bool = True
    # indices 3-5: ???
    group_id: int = field(TType.I64, index=6, default=None)
    user_id: int = field(TType.I64, default=None)
    # index 8: ???
    read_to: int = field(TType.I64, index=9)
    offline_threading_id: int = field(TType.I64, index=13)


@autospec
@dataclass
class ChatIDWrapper(ThriftObject):
    chat_id: str


@autospec
@dataclass
class OpenedThreadRequest(ThriftObject):
    unknown_i64: int = field(TType.I64, default=0)
    _chat_id: bytes = field(default=None)

    @property
    def chat_id(self) -> int:
        return int(ChatIDWrapper.from_thrift(self._chat_id).chat_id)

    @chat_id.setter
    def chat_id(self, value: int) -> None:
        self._chat_id = ChatIDWrapper(str(value)).to_thrift()



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Dict
import logging
import random
import string
import time
import json

from aiohttp import web

from mautrix.types import UserID
from mautrix.util.signed_token import verify_token
from maufbapi import AndroidState, AndroidAPI
from maufbapi.http import TwoFactorRequired, OAuthException, IncorrectPassword

from .. import user as u, puppet as pu


class PublicBridgeWebsite:
    log: logging.Logger = logging.getLogger("mau.web.public")
    app: web.Application
    secret_key: str
    shared_secret: str

    def __init__(self, shared_secret: str) -> None:
        self.app = web.Application()
        self.secret_key = "".join(random.choices(string.ascii_lowercase + string.digits, k=64))
        self.shared_secret = shared_secret
        for path in ("whoami", "login", "login/2fa", "login/check_approved", "login/approved",
                     "logout", "disconnect", "reconnect", "refresh"):
            self.app.router.add_options(f"/api/{path}", self.login_options)
        self.app.router.add_get("/api/whoami", self.status)
        self.app.router.add_post("/api/login", self.login)
        self.app.router.add_post("/api/login/2fa", self.login_2fa)
        self.app.router.add_get("/api/login/check_approved", self.login_check_approved)
        self.app.router.add_post("/api/login/approved", self.login_approved)
        self.app.router.add_post("/api/logout", self.logout)
        self.app.router.add_post("/api/disconnect", self.disconnect)
        self.app.router.add_post("/api/reconnect", self.reconnect)
        self.app.router.add_post("/api/refresh", self.refresh)

    def verify_token(self, token: str) -> Optional[UserID]:
        token = verify_token(self.secret_key, token)
        if token and token.get("expiry", 0) > int(time.time()):
            return UserID(token.get("mxid"))
        return None

    @property
    def _acao_headers(self) -> Dict[str, str]:
        return {
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Headers": "Authorization, Content-Type",
            "Access-Control-Allow-Methods": "GET, POST, OPTIONS",
        }

    @property
    def _headers(self) -> Dict[str, str]:
        return {
            **self._acao_headers,
            "Content-Type": "application/json",
        }

    async def login_options(self, _: web.Request) -> web.Response:
        return web.Response(status=200, headers=self._headers)

    async def check_token(self, request: web.Request) -> Optional['u.User']:
        try:
            token = request.headers["Authorization"]
            token = token[len("Bearer "):]
        except KeyError:
            raise web.HTTPBadRequest(text='{"error": "Missing Authorization header"}',
                                     headers=self._headers)
        except IndexError:
            raise web.HTTPBadRequest(text='{"error": "Malformed Authorization header"}',
                                     headers=self._headers)
        if self.shared_secret and token == self.shared_secret:
            try:
                user_id = request.query["user_id"]
            except KeyError:
                raise web.HTTPBadRequest(text='{"error": "Missing user_id query param"}',
                                         headers=self._headers)
        else:
            user_id = self.verify_token(token)
            if not user_id:
                raise web.HTTPForbidden(text='{"error": "Invalid token"}', headers=self._headers)

        user = await u.User.get_by_mxid(user_id)
        return user

    async def status(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)
        data = {
            "permissions": user.permission_level,
            "mxid": user.mxid,
            "facebook": None,
        }
        if user.client:
            try:
                info = await user.client.get_self()
            except Exception:
                # TODO do something?
                self.log.warning("Exception while getting self from status endpoint",
                                 exc_info=True)
            else:
                data["facebook"] = info.serialize()
                data["facebook"]["connected"] = user.is_connected
                data["facebook"]["device_displayname"] = (f"{user.state.device.manufacturer} "
                                                          f"{user.state.device.name}")
        return web.json_response(data, headers=self._acao_headers)

    async def login(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)

        try:
            data = await request.json()
        except json.JSONDecodeError:
            raise web.HTTPBadRequest(text='{"error": "Malformed JSON"}', headers=self._headers)

        try:
            email = data["email"]
            password = data["password"]
        except KeyError:
            raise web.HTTPBadRequest(text='{"error": "Missing keys"}', headers=self._headers)

        state = AndroidState()
        state.generate(user.mxid)
        api = AndroidAPI(state, log=user.log.getChild("login-api"))
        try:
            await api.mobile_config_sessionless()
            await api.login(email, password)
            await user.on_logged_in(state)
            return web.json_response({"status": "logged-in"}, headers=self._acao_headers)
        except TwoFactorRequired as e:
            user.command_status = {
                "action": "Login",
                "state": state,
                "api": api,
            }
            return web.json_response({
                "status": "two-factor",
                "error": e.data,
            }, headers=self._acao_headers)
        except OAuthException as e:
            return web.json_response({"error": str(e)}, headers=self._acao_headers)

    async def login_2fa(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)

        if not user.command_status or user.command_status["action"] != "Login":
            raise web.HTTPBadRequest(text='{"error": "No login in progress"}',
                                     headers=self._headers)

        try:
            data = await request.json()
        except json.JSONDecodeError:
            raise web.HTTPBadRequest(text='{"error": "Malformed JSON"}', headers=self._headers)

        try:
            email = data["email"]
            code = data["code"]
        except KeyError:
            raise web.HTTPBadRequest(text='{"error": "Missing keys"}', headers=self._headers)

        state: AndroidState = user.command_status["state"]
        api: AndroidAPI = user.command_status["api"]
        try:
            await api.login_2fa(email, code)
            await user.on_logged_in(state)
            return web.json_response({"status": "logged-in"}, headers=self._acao_headers)
        except IncorrectPassword:
            return web.json_response({"error": "Incorrect two-factor authentication code",
                                      "status": "incorrect-code"}, headers=self._acao_headers)
        except OAuthException as e:
            return web.json_response({"error": str(e)}, headers=self._acao_headers)

    async def login_approved(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)

        if not user.command_status or user.command_status["action"] != "Login":
            raise web.HTTPBadRequest(text='{"error": "No login in progress"}',
                                     headers=self._headers)

        state: AndroidState = user.command_status["state"]
        api: AndroidAPI = user.command_status["api"]
        try:
            await api.login_approved()
            await user.on_logged_in(state)
            return web.json_response({"status": "logged-in"}, headers=self._acao_headers)
        except OAuthException as e:
            return web.json_response({"error": str(e)}, headers=self._acao_headers)

    async def login_check_approved(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)

        if not user.command_status or user.command_status["action"] != "Login":
            raise web.HTTPBadRequest(text='{"error": "No login in progress"}',
                                     headers=self._headers)

        api: AndroidAPI = user.command_status["api"]
        approved = await api.check_approved_machine()
        return web.json_response({"approved": approved}, headers=self._acao_headers)

    async def logout(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)

        puppet = await pu.Puppet.get_by_fbid(user.fbid)
        await user.logout()
        if puppet.is_real_user:
            await puppet.switch_mxid(None, None)
        return web.json_response({}, headers=self._acao_headers)

    async def disconnect(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)
        if not user.is_connected:
            raise web.HTTPBadRequest(text='{"error": "User is not connected"}',
                                     headers=self._headers)
        user.mqtt.disconnect()
        await user.listen_task
        return web.json_response({}, headers=self._acao_headers)

    async def reconnect(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)
        if user.is_connected:
            raise web.HTTPConflict(text='{"error": "User is already connected"}',
                                   headers=self._headers)
        user.start_listen()
        return web.json_response({}, headers=self._acao_headers)

    async def refresh(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)
        await user.try_refresh()
        return web.json_response({}, headers=self._acao_headers)



"""Add double puppet base URL to puppet table

Revision ID: 3d4f3a252006
Revises: 3c59aa2a3af9
Create Date: 2020-09-18 20:06:14.910573

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3d4f3a252006'
down_revision = '3c59aa2a3af9'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('puppet', schema=None) as batch_op:
        batch_op.add_column(sa.Column('base_url', sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('puppet', schema=None) as batch_op:
        batch_op.drop_column('base_url')
    # ### end Alembic commands ###



"""Remove user agent from db

Revision ID: 97e9f7a3c070
Revises: c56c9a30b228
Create Date: 2020-05-11 23:40:17.319060

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '97e9f7a3c070'
down_revision = 'c56c9a30b228'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("user") as batch_op:
        batch_op.drop_column('user_agent')


def downgrade():
    with op.batch_alter_table("user") as batch_op:
        batch_op.add_column(sa.Column('user_agent', sa.String(length=255), autoincrement=False, nullable=True))



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, Optional, TYPE_CHECKING, ClassVar

from asyncpg import Record
from attr import dataclass

from mautrix.util.async_db import Database

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class UserContact:
    db: ClassVar[Database] = fake_db

    user: int
    contact: int
    in_community: bool

    @classmethod
    def _from_row(cls, row: Optional[Record]) -> Optional['UserContact']:
        if row is None:
            return None
        return cls(**row)

    @classmethod
    async def all(cls, user: int) -> Dict[int, 'UserContact']:
        q = ('SELECT "user", contact, in_community FROM user_contact '
             'WHERE "user"=$1')
        rows = await cls.db.fetch(q, user)
        return {up.contact: up for up in (cls._from_row(row) for row in rows)}

    @classmethod
    async def get(cls, user: int, contact: int) -> Optional['UserContact']:
        q = 'SELECT "user", contact, in_community FROM user_contact WHERE "user"=$1 AND contact=$2'
        row = await cls.db.fetchrow(q, user, contact)
        return cls._from_row(row)

    async def insert(self) -> None:
        q = 'INSERT INTO user_contact ("user", contact, in_community) VALUES ($1, $2, $3)'
        await self.db.execute(q, self.user, self.contact, self.in_community)

    async def upsert(self) -> None:
        q = ('INSERT INTO user_contact ("user", contact, in_community) VALUES ($1, $2, $3) '
             'ON CONFLICT ("user", contact) DO UPDATE SET in_community=$3')
        await self.db.execute(q, self.user, self.contact, self.in_community)

    async def delete(self) -> None:
        await self.db.execute('DELETE FROM user_contact WHERE "user"=$1 AND contact=$2',
                              self.user, self.contact)

    async def save(self) -> None:
        await self.db.execute('UPDATE user_contact SET in_community=$3 '
                              'WHERE "user"=$1 AND contact=$2',
                              self.user, self.contact, self.in_community)

    @classmethod
    async def delete_all(cls, user: int) -> None:
        await cls.db.execute('DELETE FROM user_contact WHERE "user"=$1', user)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Union
import hashlib
import time
import json

from .base import BaseAndroidAPI
from ..types import UploadResponse


class UploadAPI(BaseAndroidAPI):
    async def send_media(self, data: bytes, file_name: str, mimetype: str,
                         offline_threading_id: int, chat_id: Optional[int] = None,
                         is_group: Optional[bool] = None, timestamp: Optional[int] = None,
                         reply_to: Optional[str] = None) -> UploadResponse:
        headers = {
            **self._headers,
            "app_id": self.state.application.client_id,
            "device_id": self.state.device.uuid,
            "attempt_id": str(offline_threading_id),
            "offset": "0",
            "x-entity-length": str(len(data)),
            "x-entity-name": file_name,
            "x-entity-type": mimetype,
            "content-type": "application/octet-stream",
            "client_tags": json.dumps({"trigger": "2:thread_list:thread",
                                       "is_in_chatheads": "false"}),
            "original_timestamp": str(timestamp or int(time.time() * 1000)),
            "x-fb-rmd": "state=NO_MATCH",
            "x-msgr-region": self.state.session.region_hint,
            "x-fb-friendly-name": "post_resumable_upload_session",
        }
        if chat_id:
            headers["send_message_by_server"] = "1"
            headers["sender_fbid"] = str(self.state.session.uid)
            headers["to"] = f"tfbid_{chat_id}" if is_group else str(chat_id)
            headers["offline_threading_id"] = str(offline_threading_id)
            headers["ttl"] = "0"
            if reply_to:
                headers["replied_to_message_id"] = reply_to
        else:
            headers["send_message_by_server"] = "0"
            headers["thread_type_hint"] = "thread"
        if mimetype.startswith("image/"):
            path_type = "messenger_gif" if mimetype == "image/gif" else "messenger_image"
            headers["image_type"] = "FILE_ATTACHMENT"
        elif mimetype.startswith("video/"):
            path_type = "messenger_video"
            headers["video_type"] = "FILE_ATTACHMENT"
        elif mimetype.startswith("audio/"):
            path_type = "messenger_audio"
            headers["audio_type"] = "VOICE_MESSAGE"
            headers["is_voicemail"] = "0"
        else:
            path_type = "messenger_file"
            headers["file_type"] = "FILE_ATTACHMENT"

        file_id = hashlib.md5(data).hexdigest() + str(offline_threading_id)
        resp = await self.http.post(self.rupload_url / path_type / file_id,
                                    headers=headers, data=data)
        json_data = await self._handle_response(resp)
        self.log.trace("Upload response: %s %s", resp.status, json_data)
        return UploadResponse.deserialize(json_data)



from .client_info import RealtimeConfig, RealtimeClientInfo
from .requests import SendMessageRequest, MarkReadRequest, OpenedThreadRequest
from .message import (ThreadKey, MessageMetadata, ImageInfo, Attachment, Reaction, MentionType,
                      Mention, Message, ExtendedMessage, MessageSyncInnerEvent, MessageSyncEvent,
                      MessageSyncInnerPayload, MessageSyncPayload, BinaryData, ReadReceipt,
                      UnsendMessage, VideoInfo, AvatarChange, OwnReadReceipt, NameChange,
                      ThreadChange, ThreadChangeAction, AudioInfo, MessageSyncError, AddMember,
                      AddMemberParticipant, ExtendedAddMember, ExtendedAddMemberParticipant,
                      RemoveMember, SendMessageResponse, RegionHintPayload)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional

from attr import dataclass
import attr

from mautrix.types import SerializableAttrs


@dataclass
class UploadErrorData(SerializableAttrs['UploadErrorData']):
    retriable: bool
    type: str
    message: str


@dataclass
class UploadResponse(SerializableAttrs['UploadResponse']):
    media_id: Optional[int] = None
    err_code: Optional[str] = None
    err_str: Optional[str] = None
    is_retryable: Optional[str] = attr.ib(metadata={"json": "isRetryable"}, default=None)
    message_id: Optional[str] = None
    sent_by_server: Optional[str] = None
    success: Optional[str] = None
    debug_info: Optional[UploadErrorData] = None



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import asyncio
import logging

from mautrix.types import UserID, RoomID
from mautrix.bridge import Bridge
from mautrix.bridge.state_store.asyncpg import PgBridgeStateStore
from mautrix.util.async_db import Database

from .config import Config
from .db import upgrade_table, init as init_db
from .user import User
from .portal import Portal
from .puppet import Puppet
from .matrix import MatrixHandler
from .version import version, linkified_version
from .web import PublicBridgeWebsite


class MessengerBridge(Bridge):
    name = "mautrix-facebook"
    module = "mautrix_facebook"
    command = "python -m mautrix-facebook"
    description = "A Matrix-Facebook Messenger puppeting bridge."
    repo_url = "https://github.com/tulir/mautrix-facebook"
    real_user_content_key = "net.maunium.facebook.puppet"
    version = version
    markdown_version = linkified_version
    config_class = Config
    matrix_class = MatrixHandler

    db: Database
    config: Config
    matrix: MatrixHandler
    public_website: PublicBridgeWebsite
    state_store: PgBridgeStateStore

    periodic_reconnect_task: asyncio.Task

    def make_state_store(self) -> None:
        self.state_store = PgBridgeStateStore(self.db, self.get_puppet, self.get_double_puppet)

    def prepare_db(self) -> None:
        self.db = Database(self.config["appservice.database"], upgrade_table=upgrade_table,
                           loop=self.loop, db_args=self.config["appservice.database_opts"])
        init_db(self.db)

    def prepare_bridge(self) -> None:
        super().prepare_bridge()
        self.public_website = PublicBridgeWebsite(self.config["appservice.public.shared_secret"])
        self.az.app.add_subapp(self.config["appservice.public.prefix"], self.public_website.app)

    def prepare_stop(self) -> None:
        self.periodic_reconnect_task.cancel()
        self.log.debug("Stopping puppet syncers")
        for puppet in Puppet.by_custom_mxid.values():
            puppet.stop()
        self.log.debug("Stopping facebook listeners")
        User.shutdown = True
        for user in User.by_fbid.values():
            user.stop_listen()

    async def stop(self) -> None:
        await super().stop()
        self.log.debug("Saving user sessions")
        for user in User.by_mxid.values():
            await user.save()

    async def start(self) -> None:
        await self.db.start()
        await self.state_store.upgrade_table.upgrade(self.db.pool)
        if self.matrix.e2ee:
            self.matrix.e2ee.crypto_db.override_pool(self.db.pool)
        self.add_startup_actions(User.init_cls(self))
        self.add_startup_actions(Puppet.init_cls(self))
        Portal.init_cls(self)
        if self.config["bridge.resend_bridge_info"]:
            self.add_startup_actions(self.resend_bridge_info())
        await super().start()
        self.periodic_reconnect_task = self.loop.create_task(self._try_periodic_reconnect_loop())

    async def resend_bridge_info(self) -> None:
        self.config["bridge.resend_bridge_info"] = False
        self.config.save()
        self.log.info("Re-sending bridge info state event to all portals")
        async for portal in Portal.all():
            await portal.update_bridge_info()
        self.log.info("Finished re-sending bridge info state events")

    async def _try_periodic_reconnect_loop(self) -> None:
        try:
            await self._periodic_reconnect_loop()
        except Exception:
            self.log.exception("Fatal error in periodic reconnect loop")

    async def _periodic_reconnect_loop(self) -> None:
        log = logging.getLogger("mau.periodic_reconnect")
        always_reconnect = self.config["bridge.periodic_reconnect.always"]
        interval = self.config["bridge.periodic_reconnect.interval"]
        if interval <= 0:
            log.debug("Periodic reconnection is not enabled")
            return
        mode = self.config["bridge.periodic_reconnect.mode"].lower()
        if mode != "refresh" and mode != "reconnect":
            log.error("Invalid periodic reconnect mode '%s'", mode)
            return
        elif interval < 600:
            log.warning("Periodic reconnect interval is quite low (%d)", interval)
        log.debug("Starting periodic reconnect loop")
        while True:
            try:
                await asyncio.sleep(interval)
            except asyncio.CancelledError:
                log.debug("Periodic reconnect loop stopped")
                return
            log.info("Executing periodic reconnections")
            for user in User.by_fbid.values():
                if not user.is_connected and not always_reconnect:
                    log.debug("Not reconnecting %s: not connected", user.mxid)
                    continue
                log.debug("Executing periodic reconnect for %s", user.mxid)
                try:
                    if mode == "refresh":
                        await user.refresh()
                    elif mode == "reconnect":
                        await user.reconnect()
                except asyncio.CancelledError:
                    log.debug("Periodic reconnect loop stopped")
                    return
                except Exception:
                    log.exception("Error while reconnecting", user.mxid)

    async def get_portal(self, room_id: RoomID) -> Portal:
        return await Portal.get_by_mxid(room_id)

    async def get_puppet(self, user_id: UserID, create: bool = False) -> Puppet:
        return await Puppet.get_by_mxid(user_id, create=create)

    async def get_double_puppet(self, user_id: UserID) -> Puppet:
        return await Puppet.get_by_custom_mxid(user_id)

    async def get_user(self, user_id: UserID, create: bool = True) -> User:
        return await User.get_by_mxid(user_id, create=create)

    def is_bridge_ghost(self, user_id: UserID) -> bool:
        return bool(Puppet.get_id_from_mxid(user_id))


MessengerBridge().run()



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Union, TYPE_CHECKING
import time

from mautrix.types import (EventID, RoomID, UserID, Event, EventType, MessageEvent, StateEvent,
                           RedactionEvent, PresenceEventContent, ReceiptEvent, PresenceState,
                           ReactionEvent, ReactionEventContent, RelationType, PresenceEvent,
                           TypingEvent, TextMessageEventContent, MessageType, EncryptedEvent,
                           SingleReceiptEventContent)
from mautrix.errors import MatrixError
from mautrix.bridge import BaseMatrixHandler

from . import user as u, portal as po, puppet as pu
from .db import ThreadType, Message as DBMessage

if TYPE_CHECKING:
    from .__main__ import MessengerBridge


class MatrixHandler(BaseMatrixHandler):
    def __init__(self, bridge: 'MessengerBridge') -> None:
        prefix, suffix = bridge.config["bridge.username_template"].format(userid=":").split(":")
        homeserver = bridge.config["homeserver.domain"]
        self.user_id_prefix = f"@{prefix}"
        self.user_id_suffix = f"{suffix}:{homeserver}"
        super().__init__(bridge=bridge)

    async def send_welcome_message(self, room_id: RoomID, inviter: 'u.User') -> None:
        await super().send_welcome_message(room_id, inviter)
        if not inviter.notice_room:
            inviter.notice_room = room_id
            await inviter.save()
            await self.az.intent.send_notice(room_id, "This room has been marked as your "
                                                      "Facebook Messenger bridge notice room.")

    async def handle_puppet_invite(self, room_id: RoomID, puppet: 'pu.Puppet',
                                   invited_by: 'u.User', event_id: EventID) -> None:
        intent = puppet.default_mxid_intent
        self.log.debug(f"{invited_by.mxid} invited puppet for {puppet.fbid} to {room_id}")
        if not await invited_by.is_logged_in():
            await intent.error_and_leave(room_id, text="Please log in before inviting Facebook "
                                                       "Messenger puppets to private chats.")
            return

        portal = await po.Portal.get_by_mxid(room_id)
        if portal:
            if portal.is_direct:
                await intent.error_and_leave(room_id, text="You can not invite additional users "
                                                           "to private chats.")
                return
            # TODO add facebook inviting
            # await portal.invite_facebook(inviter, puppet)
            # await intent.join_room(room_id)
            return
        await intent.join_room(room_id)
        try:
            members = await intent.get_room_members(room_id)
        except MatrixError:
            self.log.exception(f"Failed to get member list after joining {room_id}")
            await intent.leave_room(room_id)
            return
        if len(members) > 2:
            # TODO add facebook group creating
            await intent.send_notice(room_id, "You can not invite Facebook Messenger puppets to "
                                              "multi-user rooms.")
            await intent.leave_room(room_id)
            return
        portal = await po.Portal.get_by_fbid(puppet.fbid, fb_receiver=invited_by.fbid,
                                             fb_type=ThreadType.USER)
        if portal.mxid:
            try:
                await intent.invite_user(portal.mxid, invited_by.mxid, check_cache=False)
                await intent.send_notice(room_id,
                                         text=("You already have a private chat with me "
                                               f"in room {portal.mxid}"),
                                         html=("You already have a private chat with me: "
                                               f"<a href='https://matrix.to/#/{portal.mxid}'>"
                                               "Link to room"
                                               "</a>"))
                await intent.leave_room(room_id)
                return
            except MatrixError:
                pass
        portal.mxid = room_id
        e2be_ok = await portal.check_dm_encryption()
        await portal.save()
        if e2be_ok is True:
            evt_type, content = await self.e2ee.encrypt(
                room_id, EventType.ROOM_MESSAGE,
                TextMessageEventContent(msgtype=MessageType.NOTICE,
                                        body="Portal to private chat created and end-to-bridge"
                                             " encryption enabled."))
            await intent.send_message_event(room_id, evt_type, content)
        else:
            message = "Portal to private chat created."
            if e2be_ok is False:
                message += "\n\nWarning: Failed to enable end-to-bridge encryption"
            await intent.send_notice(room_id, message)

    async def handle_invite(self, room_id: RoomID, user_id: UserID, invited_by: 'u.User',
                            event_id: EventID) -> None:
        # TODO handle puppet and user invites for group chats
        # The rest can probably be ignored
        pass

    async def handle_join(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        user = await u.User.get_by_mxid(user_id)

        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        if not user.is_whitelisted:
            await portal.main_intent.kick_user(room_id, user.mxid,
                                               "You are not whitelisted on this "
                                               "Facebook Messenger bridge.")
            return
        elif not await user.is_logged_in() and not self.config["bridge.allow_invites"]:
            await portal.main_intent.kick_user(room_id, user.mxid, "You are not logged in to this "
                                                                   "Facebook Messenger bridge.")
            return

        self.log.debug(f"{user.mxid} joined {room_id}")
        # await portal.join_matrix(user, event_id)

    async def handle_leave(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        user = await u.User.get_by_mxid(user_id, create=False)
        if not user:
            return

        await portal.handle_matrix_leave(user)

    @staticmethod
    async def handle_redaction(room_id: RoomID, user_id: UserID, event_id: EventID,
                               redaction_event_id: EventID) -> None:
        user = await u.User.get_by_mxid(user_id)
        if not user:
            return

        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        await portal.handle_matrix_redaction(user, event_id, redaction_event_id)

    @classmethod
    async def handle_reaction(cls, room_id: RoomID, user_id: UserID, event_id: EventID,
                              content: ReactionEventContent) -> None:
        if content.relates_to.rel_type != RelationType.ANNOTATION:
            cls.log.debug(f"Ignoring m.reaction event in {room_id} from {user_id} with unexpected "
                          f"relation type {content.relates_to.rel_type}")
            return
        user = await u.User.get_by_mxid(user_id)
        if not user:
            return

        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        await portal.handle_matrix_reaction(user, event_id, content.relates_to.event_id,
                                            content.relates_to.key)

    async def handle_presence(self, user_id: UserID, info: PresenceEventContent) -> None:
        if not self.config["bridge.presence"]:
            return
        # user = await u.User.get_by_mxid(user_id, create=False)
        # if user and user.mqtt:
        #     user.log.debug(f"Setting foreground status to {info.presence == PresenceState.ONLINE}")
        #     user.mqtt.set_foreground(info.presence == PresenceState.ONLINE)

    @staticmethod
    async def handle_typing(room_id: RoomID, typing: List[UserID]) -> None:
        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        # FIXME
        # users = [await u.User.get_by_mxid(mxid, create=False) for mxid in typing]
        # await portal.handle_matrix_typing({user for user in users
        #                                    if user is not None})

    async def handle_read_receipt(self, user: 'u.User', portal: 'po.Portal', event_id: EventID,
                                  data: SingleReceiptEventContent) -> None:
        timestamp = data.get("ts", int(time.time() * 1000))
        message = await DBMessage.get_by_mxid(event_id, portal.mxid)
        await user.mqtt.mark_read(portal.fbid, portal.fb_type != ThreadType.USER,
                                  read_to=message.timestamp if message else timestamp)

    def filter_matrix_event(self, evt: Event) -> bool:
        if isinstance(evt, (ReceiptEvent, TypingEvent, PresenceEvent)):
            return False
        elif not isinstance(evt, (ReactionEvent, RedactionEvent, MessageEvent, StateEvent,
                                  EncryptedEvent)):
            return True
        return (evt.sender == self.az.bot_mxid
                or pu.Puppet.get_id_from_mxid(evt.sender) is not None)

    async def handle_ephemeral_event(self, evt: Union[ReceiptEvent, PresenceEvent, TypingEvent]
                                     ) -> None:
        if evt.type == EventType.PRESENCE:
            await self.handle_presence(evt.sender, evt.content)
        elif evt.type == EventType.TYPING:
            await self.handle_typing(evt.room_id, evt.content.user_ids)
        elif evt.type == EventType.RECEIPT:
            await self.handle_receipt(evt)

    async def handle_event(self, evt: Event) -> None:
        if evt.type == EventType.ROOM_REDACTION:
            evt: RedactionEvent
            await self.handle_redaction(evt.room_id, evt.sender, evt.redacts, evt.event_id)
        elif evt.type == EventType.REACTION:
            evt: ReactionEvent
            await self.handle_reaction(evt.room_id, evt.sender, evt.event_id, evt.content)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, List, TYPE_CHECKING, ClassVar

from asyncpg import Record
from attr import dataclass
from yarl import URL

from mautrix.types import UserID, SyncToken, ContentURI
from mautrix.util.async_db import Database

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class Puppet:
    db: ClassVar[Database] = fake_db

    fbid: int
    name: Optional[str]
    photo_id: Optional[str]
    photo_mxc: Optional[ContentURI]
    name_set: bool
    avatar_set: bool
    is_registered: bool

    custom_mxid: Optional[UserID]
    access_token: Optional[str]
    next_batch: Optional[SyncToken]
    base_url: Optional[URL]

    @classmethod
    def _from_row(cls, row: Optional[Record]) -> Optional['Puppet']:
        if row is None:
            return None
        data = {**row}
        base_url = data.pop("base_url", None)
        return cls(**data, base_url=URL(base_url) if base_url else None)

    @classmethod
    async def get_by_fbid(cls, fbid: int) -> Optional['Puppet']:
        q = ("SELECT fbid, name, photo_id, photo_mxc, name_set, avatar_set, is_registered, "
             "       custom_mxid, access_token, next_batch, base_url "
             "FROM puppet WHERE fbid=$1")
        row = await cls.db.fetchrow(q, fbid)
        return cls._from_row(row)

    @classmethod
    async def get_by_name(cls, name: str) -> Optional['Puppet']:
        q = ("SELECT fbid, name, photo_id, photo_mxc, name_set, avatar_set, is_registered, "
             "       custom_mxid, access_token, next_batch, base_url "
             "FROM puppet WHERE name=$1")
        row = await cls.db.fetchrow(q, name)
        return cls._from_row(row)

    @classmethod
    async def get_by_custom_mxid(cls, mxid: UserID) -> Optional['Puppet']:
        q = ("SELECT fbid, name, photo_id, photo_mxc, name_set, avatar_set, is_registered, "
             "       custom_mxid, access_token, next_batch, base_url "
             "FROM puppet WHERE custom_mxid=$1")
        row = await cls.db.fetchrow(q, mxid)
        return cls._from_row(row)

    @classmethod
    async def get_all_with_custom_mxid(cls) -> List['Puppet']:
        q = ("SELECT fbid, name, photo_id, photo_mxc, name_set, avatar_set, is_registered, "
             "       custom_mxid, access_token, next_batch, base_url "
             "FROM puppet WHERE custom_mxid<>''")
        rows = await cls.db.fetch(q)
        return [cls._from_row(row) for row in rows]

    async def insert(self) -> None:
        q = ("INSERT INTO puppet (fbid, name, photo_id, photo_mxc, name_set, avatar_set, "
             "                    is_registered, custom_mxid, access_token, next_batch, base_url) "
             "VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11)")
        await self.db.execute(q, self.fbid, self.name, self.photo_id, self.photo_mxc,
                              self.name_set, self.avatar_set, self.is_registered, self.custom_mxid,
                              self.access_token, self.next_batch,
                              str(self.base_url) if self.base_url else None)

    async def delete(self) -> None:
        q = "DELETE FROM puppet WHERE fbid=$1"
        await self.db.execute(q, self.fbid)

    async def save(self) -> None:
        q = ('UPDATE puppet SET name=$2, photo_id=$3, photo_mxc=$4, name_set=$5, avatar_set=$6, '
             '                  is_registered=$7, custom_mxid=$8, access_token=$9, next_batch=$10,'
             '                  base_url=$11 '
             'WHERE fbid=$1')
        await self.db.execute(q, self.fbid, self.name, self.photo_id, self.photo_mxc,
                              self.name_set, self.avatar_set, self.is_registered, self.custom_mxid,
                              self.access_token, self.next_batch,
                              str(self.base_url) if self.base_url else None)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union, Optional, Any, Dict, Awaitable, Type, List, TypeVar, Callable
from collections import defaultdict
from socket import socket, error as SocketError
import urllib.request
import logging
import asyncio
import random
import zlib
import time
import json

import paho.mqtt.client
from paho.mqtt.client import MQTTMessage, WebsocketConnectionError
from yarl import URL
from mautrix.util.logging import TraceLogger

from ..state import AndroidState
from ..types import (MessageSyncPayload, RealtimeConfig, RealtimeClientInfo, SendMessageRequest,
                     MarkReadRequest, OpenedThreadRequest, SendMessageResponse, RegionHintPayload)
from ..types.mqtt import Mention
from ..thrift import ThriftReader, ThriftObject
from .otclient import MQTToTClient
from .subscription import RealtimeTopic, topic_map
from .events import Connect, Disconnect

try:
    import socks
except ImportError:
    socks = None

T = TypeVar('T')


# TODO add some custom stuff in these?
class MQTTNotLoggedIn(Exception):
    pass


class MQTTNotConnected(Exception):
    pass


class AndroidMQTT:
    _loop: asyncio.AbstractEventLoop
    _client: MQTToTClient
    log: TraceLogger
    state: AndroidState
    seq_id: Optional[int]
    seq_id_update_callback: Optional[Callable[[int], None]]
    region_hint_callback: Optional[Callable[[str], None]]
    _opened_thread: Optional[int]
    _publish_waiters: Dict[int, asyncio.Future]
    _response_waiters: Dict[RealtimeTopic, asyncio.Future]
    _response_waiter_locks: Dict[RealtimeTopic, asyncio.Lock]
    _disconnect_error: Optional[Exception]
    _event_handlers: Dict[Type[T], List[Callable[[T], Awaitable[None]]]]

    # region Initialization

    def __init__(self, state: AndroidState, loop: Optional[asyncio.AbstractEventLoop] = None,
                 log: Optional[TraceLogger] = None) -> None:
        self.seq_id = None
        self.seq_id_update_callback = None
        self.region_hint_callback = None
        self._opened_thread = None
        self._publish_waiters = {}
        self._response_waiters = {}
        self._disconnect_error = None
        self._response_waiter_locks = defaultdict(lambda: asyncio.Lock())
        self._event_handlers = defaultdict(lambda: [])
        self.log = log or logging.getLogger("mauigpapi.mqtt")
        self._loop = loop or asyncio.get_event_loop()
        self.state = state
        self._client = MQTToTClient(
            client_id=self._form_client_id(),
            clean_session=True,
            protocol=paho.mqtt.client.MQTTv31,
            transport="tcp",
        )
        try:
            http_proxy = urllib.request.getproxies()["http"]
        except KeyError:
            http_proxy = None
        if http_proxy and socks and URL:
            proxy_url = URL(http_proxy)
            proxy_type = {
                "http": socks.HTTP,
                "https": socks.HTTP,
                "socks": socks.SOCKS5,
                "socks5": socks.SOCKS5,
                "socks4": socks.SOCKS4,
            }[proxy_url.scheme]
            self._client.proxy_set(proxy_type=proxy_type, proxy_addr=proxy_url.host,
                                   proxy_port=proxy_url.port, proxy_username=proxy_url.user,
                                   proxy_password=proxy_url.password)
        self._client.enable_logger()
        self._client.tls_set()
        # mqtt.max_inflight_messages_set(20)  # The rest will get queued
        # mqtt.max_queued_messages_set(0)  # Unlimited messages can be queued
        # mqtt.message_retry_set(20)  # Retry sending for at least 20 seconds
        # mqtt.reconnect_delay_set(min_delay=1, max_delay=120)
        self._client.connect_async("edge-mqtt.facebook.com", 443, keepalive=60)
        self._client.on_message = self._on_message_handler
        self._client.on_publish = self._on_publish_handler
        self._client.on_connect = self._on_connect_handler
        # self._client.on_disconnect = self._on_disconnect_handler
        self._client.on_socket_open = self._on_socket_open
        self._client.on_socket_close = self._on_socket_close
        self._client.on_socket_register_write = self._on_socket_register_write
        self._client.on_socket_unregister_write = self._on_socket_unregister_write

    def _form_client_id(self) -> bytes:
        # subscribe_topics = ["/t_p", "/t_assist_rp", "/t_rtc", "/webrtc_response",
        #                     RealtimeTopic.MESSAGE_SYNC, "/pp", "/webrtc",
        #                     "/quick_promotion_refresh", "/t_omnistore_sync_low_pri",
        #                     "/get_media_resp", "/t_dr_response", "/t_omnistore_sync", "/t_push",
        #                     "/t_thread_typing", "/ixt_trigger", "/rs_resp",
        #                     RealtimeTopic.REGION_HINT, "/t_tn", "/sr_res", "/t_tp", "/t_sp",
        #                     "/ls_resp", "/t_rtc_multi",  # RealtimeTopic.SEND_MESSAGE_RESP,
        #                     ]
        subscribe_topics = [RealtimeTopic.MESSAGE_SYNC, RealtimeTopic.REGION_HINT,
                            RealtimeTopic.SEND_MESSAGE_RESP,
                            RealtimeTopic.MARK_THREAD_READ_RESPONSE]
        topic_ids = [int(topic.encoded if isinstance(topic, RealtimeTopic) else topic_map[topic])
                     for topic in subscribe_topics]
        cfg = RealtimeConfig(
            client_identifier=self.state.device.uuid[:20],
            client_info=RealtimeClientInfo(
                user_id=self.state.session.uid,
                user_agent=self.state.user_agent_meta,
                client_capabilities=0b100001110110111,
                endpoint_capabilities=0b1011010,
                publish_format=1,
                no_automatic_foreground=True,
                make_user_available_in_foreground=True,
                device_id=self.state.device.uuid,
                is_initially_foreground=True,
                network_type=1,
                network_subtype=0,
                client_mqtt_session_id=int(time.time() * 1000) & 0xffffffff,
                subscribe_topics=topic_ids,
                client_type="",
                app_id=int(self.state.application.client_id),
                region_preference=self.state.session.region_hint or "ODN",
                device_secret="",
                client_stack=4,
                yet_another_unknown=7,
            ),
            password=self.state.session.access_token,
            app_specific_info={
                "ls_sv": str(self.state.application.version_id),
                "ls_fdid": self.state.device.uuid,
            },
        )
        return zlib.compress(cfg.to_thrift(), level=9)

    # endregion

    def _on_socket_open(self, client: MQTToTClient, _: Any, sock: socket) -> None:
        self._loop.add_reader(sock, client.loop_read)

    def _on_socket_close(self, client: MQTToTClient, _: Any, sock: socket) -> None:
        self._loop.remove_reader(sock)

    def _on_socket_register_write(self, client: MQTToTClient, _: Any, sock: socket) -> None:
        self._loop.add_writer(sock, client.loop_write)

    def _on_socket_unregister_write(self, client: MQTToTClient, _: Any, sock: socket) -> None:
        self._loop.remove_writer(sock)

    def _on_connect_handler(self, client: MQTToTClient, _: Any, flags: Dict[str, Any], rc: int
                            ) -> None:
        if rc != 0:
            if rc == paho.mqtt.client.MQTT_ERR_INVAL:
                self.log.error("MQTT connection error, regenerating client ID")
                self._client.set_client_id(self._form_client_id())
            else:
                err = paho.mqtt.client.connack_string(rc)
                self.log.error("MQTT Connection Error: %s (%d)", err, rc)
            return

        self._loop.create_task(self._post_connect())

    async def _post_connect(self) -> None:
        self._opened_thread = None
        self.log.debug("Re-creating sync queue after reconnect")
        await self._dispatch(Connect())
        await self.publish("/ls_req", {
            "label": "1",
            "payload": json.dumps({
                "app_state": 1,
                "request_id": "android_request_id",
            }),
            "version": str(self.state.application.version_id),
        })
        await self.publish(RealtimeTopic.SYNC_CREATE_QUEUE, {
            "initial_titan_sequence_id": self.seq_id,
            "delta_batch_size": 125,
            "device_params": {
                "image_sizes": {
                    "0": "4096x4096",
                    "4": "358x358",
                    "1": "750x750",
                    "2": "481x481",
                    "3": "358x358"
                },
                "animated_image_format": "WEBP,GIF",
                "animated_image_sizes": {
                    "0": "4096x4096",
                    "4": "358x358",
                    "1": "750x750",
                    "2": "481x481",
                    "3": "358x358"
                },
                "thread_theme_background_sizes": {
                    "0": "2048x2048"
                },
                "thread_theme_icon_sizes": {
                    "1": "138x138",
                    "3": "66x66"
                },
                "thread_theme_reaction_sizes": {
                    "1": "83x83",
                    "3": "39x39"
                }
            },
            "entity_fbid": self.state.session.uid,
            "sync_api_version": 10,
            "queue_params": {
                "client_delta_sync_bitmask": "B/p8Ym/r2YAFf7PNgA",
                "graphql_query_hashes": {
                    "xma_query_id": "3257579454369025"
                },
                "graphql_query_params": {
                    "3257579454369025": {
                        "xma_id": "<ID>",
                        "small_preview_width": 716,
                        "small_preview_height": 358,
                        "large_preview_width": 1500,
                        "large_preview_height": 750,
                        "full_screen_width": 4096,
                        "full_screen_height": 4096,
                        "blur": 0,
                        "nt_context": {
                            "styles_id": "7d328425a4dfa3aa76b1310fa8dc30bf",
                            "pixel_ratio": 3
                        },
                        "use_oss_id": True
                    }
                }
            }
        })

    def _on_publish_handler(self, client: MQTToTClient, _: Any, mid: int) -> None:
        try:
            waiter = self._publish_waiters[mid]
        except KeyError:
            return
        waiter.set_result(None)

    # region Incoming event parsing

    def _update_seq_id(self, msp: MessageSyncPayload) -> None:
        if msp.last_seq_id and msp.last_seq_id > self.seq_id:
            self.seq_id = msp.last_seq_id
            self.seq_id_update_callback(self.seq_id)

    def _on_message_sync(self, payload: bytes) -> None:
        parsed = MessageSyncPayload.from_thrift(payload)
        self._update_seq_id(parsed)
        if parsed.error:
            self._loop.create_task(self._dispatch(parsed.error))
        for item in parsed.items:
            for event in item.get_parts():
                self._loop.create_task(self._dispatch(event))

    def _on_region_hint(self, payload: bytes) -> None:
        rhp = RegionHintPayload.from_thrift(payload)
        if self.region_hint_callback:
            self.region_hint_callback(rhp.region_hint.code)

    def _on_message_handler(self, client: MQTToTClient, _: Any, message: MQTTMessage) -> None:
        try:
            is_compressed = message.payload.startswith(b"x\xda")
            if is_compressed:
                message.payload = zlib.decompress(message.payload)
            topic = RealtimeTopic.decode(message.topic)
            _, message.payload = message.payload.split(b"\x00", 1)
            if topic == RealtimeTopic.MESSAGE_SYNC:
                self._on_message_sync(message.payload)
            elif topic == RealtimeTopic.REGION_HINT:
                self._on_region_hint(message.payload)
            else:
                try:
                    waiter = self._response_waiters.pop(topic)
                except KeyError:
                    self.log.debug("No handler for MQTT message in %s: %s",
                                   topic, message.payload)
                    ThriftReader(message.payload).pretty_print()
                else:
                    waiter.set_result(message)
        except Exception:
            self.log.exception("Error in incoming MQTT message handler")
            self.log.trace("Errored MQTT payload: %s", message.payload)

    # endregion

    async def _reconnect(self) -> None:
        try:
            self.log.trace("Trying to reconnect to MQTT")
            self._client.reconnect()
        except (SocketError, OSError, WebsocketConnectionError) as e:
            raise MQTTNotLoggedIn("MQTT reconnection failed") from e

    def add_event_handler(self, evt_type: Type[T], handler: Callable[[T], Awaitable[None]]
                          ) -> None:
        self._event_handlers[evt_type].append(handler)

    async def _dispatch(self, evt: T) -> None:
        for handler in self._event_handlers[type(evt)]:
            try:
                await handler(evt)
            except Exception:
                self.log.exception(f"Error in {type(evt).__name__} handler")
                self.log.trace("Event content: %s", evt)

    def disconnect(self) -> None:
        self._client.disconnect()

    async def listen(self, seq_id: int, retry_limit: int = 5) -> None:
        self.seq_id = seq_id

        self.log.debug("Connecting to Messenger MQTT")
        await self._reconnect()
        connection_retries = 0

        while True:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.disconnect()
                # this might not be necessary
                self._client.loop_misc()
                break
            rc = self._client.loop_misc()

            # If disconnect() has been called
            # Beware, internal API, may have to change this to something more stable!
            if self._client._state == paho.mqtt.client.mqtt_cs_disconnecting:
                break  # Stop listening

            if rc != paho.mqtt.client.MQTT_ERR_SUCCESS:
                # If known/expected error
                if rc == paho.mqtt.client.MQTT_ERR_CONN_LOST:
                    await self._dispatch(Disconnect(reason="Connection lost, retrying"))
                elif rc == paho.mqtt.client.MQTT_ERR_NOMEM:
                    # This error is wrongly classified
                    # See https://github.com/eclipse/paho.mqtt.python/issues/340
                    await self._dispatch(Disconnect(reason="Connection lost, retrying"))
                elif rc == paho.mqtt.client.MQTT_ERR_CONN_REFUSED:
                    raise MQTTNotLoggedIn("MQTT connection refused")
                elif rc == paho.mqtt.client.MQTT_ERR_NO_CONN:
                    if connection_retries > retry_limit:
                        raise MQTTNotConnected(f"Connection failed {connection_retries} times")
                    sleep = connection_retries * 2
                    await self._dispatch(Disconnect(reason="MQTT Error: no connection, retrying "
                                                           f"in {connection_retries} seconds"))
                    await asyncio.sleep(sleep)
                else:
                    err = paho.mqtt.client.error_string(rc)
                    self.log.error("MQTT Error: %s", err)
                    await self._dispatch(Disconnect(reason=f"MQTT Error: {err}, retrying"))

                await self._reconnect()
                connection_retries += 1
            else:
                connection_retries = 0
        if self._disconnect_error:
            self.log.info("disconnect_error is set, raising and clearing variable")
            err = self._disconnect_error
            self._disconnect_error = None
            raise err

    # region Basic outgoing MQTT

    def publish(self, topic: Union[RealtimeTopic, str],
                payload: Union[str, bytes, dict, ThriftObject],
                prefix: bytes = b"") -> asyncio.Future:
        if isinstance(payload, dict):
            payload = json.dumps(payload)
        if isinstance(payload, str):
            payload = payload.encode("utf-8")
        if isinstance(payload, ThriftObject):
            payload = payload.to_thrift()
        payload = zlib.compress(prefix + payload, level=9)
        info = self._client.publish(topic.encoded if isinstance(topic, RealtimeTopic) else topic,
                                    payload, qos=1)
        fut = asyncio.Future()
        self._publish_waiters[info.mid] = fut
        return fut

    async def request(self, topic: RealtimeTopic, response: RealtimeTopic,
                      payload: Union[str, bytes, dict, ThriftObject], prefix: bytes = b""
                      ) -> MQTTMessage:
        async with self._response_waiter_locks[response]:
            fut = asyncio.Future()
            self._response_waiters[response] = fut
            await self.publish(topic, payload, prefix)
            return await fut

    @staticmethod
    def generate_offline_threading_id() -> int:
        rand = format(int(random.random() * 4294967295), "022b")[-22:]
        return int(f"{int(time.time() * 1000):b}{rand}", 2)

    async def send_message(self, target: int, is_group: bool, message: str = "",
                           offline_threading_id: Optional[int] = None, media_ids: List[int] = None,
                           mentions: Optional[List[Mention]] = None, reply_to: Optional[str] = None
                           ) -> SendMessageResponse:
        if not offline_threading_id:
            offline_threading_id = self.generate_offline_threading_id()
        req = SendMessageRequest(chat_id=f"tfbid_{target}" if is_group else str(target),
                                 message=message, offline_threading_id=offline_threading_id,
                                 sender_id=self.state.session.uid, reply_to=reply_to,
                                 media_ids=[str(i) for i in media_ids] if media_ids else None,
                                 flags={"is_in_chatheads": "false",
                                        "trigger": "2:thread_list:thread"},
                                 tid2=self.generate_offline_threading_id())
        if mentions:
            req.extra_metadata = {"prng": json.dumps([mention.serialize() for mention in mentions],
                                                     separators=(',', ':'))}
        await self.opened_thread(target)
        self.log.trace("Send message request: %s", req)
        resp = await self.request(RealtimeTopic.SEND_MESSAGE, RealtimeTopic.SEND_MESSAGE_RESP, req,
                                  prefix=b"\x18\x00\x00")
        self.log.trace("Send message response: %s", repr(resp.payload))
        return SendMessageResponse.from_thrift(resp.payload)

    async def opened_thread(self, target: int) -> None:
        if self._opened_thread == target:
            return
        self._opened_thread = target
        req = OpenedThreadRequest()
        req.chat_id = target
        self.log.trace("Opened thread request: %s", req)
        await self.publish(RealtimeTopic.OPENED_THREAD, req)

    async def mark_read(self, target: int, is_group: bool, read_to: int,
                        offline_threading_id: Optional[int] = None) -> None:
        if not offline_threading_id:
            offline_threading_id = self.generate_offline_threading_id()
        req = MarkReadRequest(read_to=read_to, offline_threading_id=offline_threading_id)
        if is_group:
            req.group_id = target
        else:
            req.user_id = target
        await self.opened_thread(target)
        self.log.trace("Mark read request: %s", req)
        resp = await self.request(RealtimeTopic.MARK_THREAD_READ,
                                  RealtimeTopic.MARK_THREAD_READ_RESPONSE,
                                  req, prefix=b"\x00")
        self.log.trace("Mark read response: %s", repr(resp.payload))

    # endregion



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Tuple, List, Optional, Match, Union
from html import escape
import re

from mautrix.types import TextMessageEventContent, Format, MessageType
from maufbapi.types import graphql, mqtt

from .. import puppet as pu, user as u

_START = r"^|\s"
_END = r"$|\s"
_TEXT_NO_SURROUNDING_SPACE = r"(?:[^\s].*?[^\s])|[^\s]"
COMMON_REGEX = re.compile(rf"({_START})([_~*])({_TEXT_NO_SURROUNDING_SPACE})\2({_END})")
INLINE_CODE_REGEX = re.compile(rf"({_START})(`)(.+?)`({_END})")
MENTION_REGEX = re.compile(r"@([0-9]{1,15})\u2063(.+)\u2063")

tags = {
    "_": "em",
    "*": "strong",
    "~": "del",
    "`": "code"
}


def _handle_match(html: str, match: Match, nested: bool) -> Tuple[str, int]:
    start, end = match.start(), match.end()
    prefix, sigil, text, suffix = match.groups()
    if nested:
        text = _convert_formatting(text)
    tag = tags[sigil]
    # We don't want to include the whitespace suffix length, as that could be used as the
    # whitespace prefix right after this formatting block.
    pos = start + len(prefix) + (2 * len(tag) + 5) + len(text)
    html = (f"{html[:start]}{prefix}"
            f"<{tag}>{text}</{tag}>"
            f"{suffix}{html[end:]}")
    return html, pos


def _convert_formatting(html: str) -> str:
    pos = 0
    while pos < len(html):
        i_match = INLINE_CODE_REGEX.search(html, pos)
        c_match = COMMON_REGEX.search(html, pos)
        if i_match and c_match:
            match = min(i_match, c_match, key=lambda match: match.start())
        else:
            match = i_match or c_match

        if match:
            html, pos = _handle_match(html, match, nested=match != i_match)
        else:
            break
    return html


def _handle_blockquote(output: List[str], blockquote: bool, line: str) -> Tuple[bool, str]:
    if not blockquote and line.startswith("&gt; "):
        line = line[len("&gt; "):]
        output.append("<blockquote>")
        blockquote = True
    elif blockquote:
        if line.startswith("&gt;"):
            line = line[len("&gt;"):]
            if line.startswith(" "):
                line = line[1:]
        else:
            output.append("</blockquote>")
            blockquote = False
    return blockquote, line


OptStr = Optional[str]


def _handle_codeblock_pre(output: List[str], codeblock: bool, line: str
                          ) -> Tuple[bool, str, Tuple[OptStr, OptStr, OptStr]]:
    cb = line.find("```")
    cb_lang = None
    cb_content = None
    post_cb_content = None
    if cb != -1:
        if not codeblock:
            cb_lang = line[cb + 3:]
            if "```" in cb_lang:
                end = cb_lang.index("```")
                cb_content = cb_lang[:end]
                post_cb_content = cb_lang[end + 3:]
                cb_lang = ""
            else:
                codeblock = True
            line = line[:cb]
        else:
            output.append("</code></pre>")
            codeblock = False
            line = line[cb + 3:]
    return codeblock, line, (cb_lang, cb_content, post_cb_content)


def _handle_codeblock_post(output: List[str], cb_lang: OptStr, cb_content: OptStr,
                           post_cb_content: OptStr) -> None:
    if cb_lang is not None:
        if cb_lang:
            output.append("<pre><code>")
        else:
            output.append(f"<pre><code class=\"{cb_lang}\">")
        if cb_content:
            output.append(cb_content)
            output.append("</code></pre>")
            output.append(_convert_formatting(post_cb_content))


async def facebook_to_matrix(msg: Union[graphql.MessageText, mqtt.Message]) -> TextMessageEventContent:
    if isinstance(msg, mqtt.Message):
        text = msg.text
        mentions = msg.mentions
    elif isinstance(msg, graphql.MessageText):
        text = msg.text
        mentions = msg.ranges
    else:
        raise ValueError(f"Unsupported Facebook message type {type(msg).__name__}")
    text = text or ""
    content = TextMessageEventContent(msgtype=MessageType.TEXT, body=text)
    mention_user_ids = []
    for m in reversed(mentions):
        original = text[m.offset:m.offset + m.length]
        if len(original) > 0 and original[0] == "@":
            original = original[1:]
        mention_user_ids.append(int(m.user_id))
        text = f"{text[:m.offset]}@{m.user_id}\u2063{original}\u2063{text[m.offset + m.length:]}"
    html = escape(text)
    output = []
    if html:
        codeblock = False
        blockquote = False
        line: str
        lines = html.split("\n")
        for i, line in enumerate(lines):
            blockquote, line = _handle_blockquote(output, blockquote, line)
            codeblock, line, post_args = _handle_codeblock_pre(output, codeblock, line)
            output.append(_convert_formatting(line))
            if i != len(lines) - 1:
                output.append("<br/>")
            _handle_codeblock_post(output, *post_args)
    html = "".join(output)

    mention_user_map = {}
    for fbid in mention_user_ids:
        user = await u.User.get_by_fbid(fbid)
        if user:
            mention_user_map[fbid] = user.mxid
        else:
            puppet = await pu.Puppet.get_by_fbid(fbid, create=False)
            mention_user_map[fbid] = puppet.mxid if puppet else None

    def _mention_replacer(match: Match) -> str:
        mxid = mention_user_map[int(match.group(1))]
        if not mxid:
            return match.group(2)
        return f"<a href=\"https://matrix.to/#/{mxid}\">{match.group(2)}</a>"

    html = MENTION_REGEX.sub(_mention_replacer, html)
    if html != escape(content.body).replace("\n", "<br/>"):
        content.format = Format.HTML
        content.formatted_body = html
    return content



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from asyncpg import (Connection, UndefinedObjectError, DuplicateObjectError,
                     ForeignKeyViolationError)
from . import upgrade_table

legacy_exist_query = ("SELECT EXISTS(SELECT FROM information_schema.tables "
                      "              WHERE table_name='alembic_version')")
legacy_version_query = "SELECT version_num FROM alembic_version"
last_legacy_version = "f91274813e8c"
legacy_renamed_query = ("SELECT EXISTS(SELECT FROM information_schema.tables "
                        "              WHERE table_name='legacy_contact')")
new_tables_created_query = ("SELECT EXISTS(SELECT FROM information_schema.tables "
                            "              WHERE table_name='user_contact')")


@upgrade_table.register(description="Initial asyncpg revision", transaction=False)
async def upgrade_v1(conn: Connection) -> None:
    try:
        async with conn.transaction():
            await conn.execute("CREATE TYPE threadtype AS ENUM "
                               "('USER', 'GROUP', 'PAGE', 'UNKNOWN')")
    except DuplicateObjectError:
        pass

    is_legacy = await conn.fetchval(legacy_exist_query)
    if is_legacy:
        legacy_version = await conn.fetchval(legacy_version_query)
        if legacy_version != last_legacy_version:
            raise RuntimeError("Legacy database is not on last version. Please upgrade the old "
                               "database with alembic or drop it completely first.")
        already_renamed = await conn.fetchval(legacy_renamed_query)
        if not already_renamed:
            async with conn.transaction():
                await rename_legacy_tables(conn)
        new_created = await conn.fetchval(new_tables_created_query)
        if not new_created:
            async with conn.transaction():
                await create_v1_tables(conn)
        async with conn.transaction():
            await migrate_legacy_data(conn)
    else:
        await create_v1_tables(conn)


async def create_v1_tables(conn: Connection) -> None:
    await conn.execute("""CREATE TABLE "user" (
        mxid        TEXT PRIMARY KEY,
        fbid        BIGINT UNIQUE,
        state       jsonb,
        notice_room TEXT
    )""")
    await conn.execute("""CREATE TABLE portal (
        fbid        BIGINT,
        fb_receiver BIGINT,
        fb_type     threadtype NOT NULL,
        mxid        TEXT UNIQUE,
        name        TEXT,
        photo_id    TEXT,
        avatar_url  TEXT,
        encrypted   BOOLEAN NOT NULL DEFAULT false,
        PRIMARY KEY (fbid, fb_receiver)
    )""")
    await conn.execute("""CREATE TABLE puppet (
        fbid      BIGINT PRIMARY KEY,
        name      TEXT,
        photo_id  TEXT,
        photo_mxc TEXT,

        name_set      BOOLEAN NOT NULL DEFAULT false,
        avatar_set    BOOLEAN NOT NULL DEFAULT false,
        is_registered BOOLEAN NOT NULL DEFAULT false,

        custom_mxid  TEXT,
        access_token TEXT,
        next_batch   TEXT,
        base_url     TEXT
    )""")
    await conn.execute("""CREATE TABLE message (
        mxid        TEXT,
        mx_room     TEXT,
        fbid        TEXT,
        fb_receiver BIGINT,
        index       SMALLINT,
        fb_chat     BIGINT,
        timestamp   BIGINT,
        PRIMARY KEY (fbid, fb_receiver, index),
        FOREIGN KEY (fb_chat, fb_receiver) REFERENCES portal(fbid, fb_receiver)
            ON UPDATE CASCADE ON DELETE CASCADE,
        UNIQUE (mxid, mx_room)
    )""")
    await conn.execute("""CREATE TABLE reaction (
        mxid        TEXT,
        mx_room     TEXT,
        fb_msgid    TEXT,
        fb_receiver BIGINT,
        fb_sender   BIGINT,
        reaction    TEXT,
        PRIMARY KEY (fb_msgid, fb_receiver, fb_sender),
        UNIQUE (mxid, mx_room)
    )""")
    await conn.execute("""CREATE TABLE user_portal (
        "user"          BIGINT,
        portal          BIGINT,
        portal_receiver BIGINT,
        in_community    BOOLEAN DEFAULT false,
        FOREIGN KEY (portal, portal_receiver) REFERENCES portal(fbid, fb_receiver)
            ON UPDATE CASCADE ON DELETE CASCADE,
        FOREIGN KEY ("user") REFERENCES "user"(fbid) ON UPDATE CASCADE ON DELETE CASCADE,
        PRIMARY KEY ("user", portal, portal_receiver)
    )""")
    await conn.execute("""CREATE TABLE user_contact (
        "user"       BIGINT,
        contact      BIGINT,
        in_community BOOLEAN DEFAULT false,
        FOREIGN KEY (contact) REFERENCES puppet(fbid)  ON UPDATE CASCADE ON DELETE CASCADE,
        FOREIGN KEY ("user") REFERENCES "user"(fbid) ON UPDATE CASCADE ON DELETE CASCADE,
        PRIMARY KEY ("user", contact)
    )""")


async def rename_legacy_tables(conn: Connection) -> None:
    await conn.execute("ALTER TABLE mx_user_profile RENAME TO legacy_mx_user_profile")
    await conn.execute("ALTER TABLE mx_room_state RENAME TO legacy_mx_room_state")
    try:
        async with conn.transaction():
            await conn.execute("ALTER TYPE membership RENAME TO legacy_membership")
    except UndefinedObjectError:
        pass

    await conn.execute("ALTER TABLE message RENAME TO legacy_message")
    await conn.execute("ALTER TABLE portal RENAME TO legacy_portal")
    await conn.execute("ALTER TABLE puppet RENAME TO legacy_puppet")
    await conn.execute("ALTER TABLE reaction RENAME TO legacy_reaction")
    await conn.execute('ALTER TABLE "user" RENAME TO legacy_user')
    await conn.execute("ALTER TABLE user_portal RENAME TO legacy_user_portal")
    await conn.execute("ALTER TABLE contact RENAME TO legacy_contact")


async def migrate_legacy_data(conn: Connection) -> None:
    await conn.execute('INSERT INTO "user" (mxid, fbid, notice_room) '
                       "SELECT mxid, fbid::bigint, notice_room FROM legacy_user")
    await conn.execute(
        "INSERT INTO portal (fbid, fb_receiver, fb_type, mxid, name, photo_id, encrypted) "
        "SELECT fbid::bigint, fb_receiver::bigint, fb_type::threadtype, mxid, name, photo_id, "
        "       encrypted "
        "FROM legacy_portal"
    )
    await conn.execute(
        "INSERT INTO puppet (fbid, name, photo_id, name_set, avatar_set, is_registered, "
        "                    custom_mxid, access_token, next_batch, base_url) "
        "SELECT fbid::bigint, name, photo_id, name_set, avatar_set, matrix_registered, "
        "       custom_mxid, access_token, next_batch, base_url "
        "FROM legacy_puppet WHERE fbid ~ '^[0-9]+$'"
    )
    await conn.execute(
        "INSERT INTO message (mxid, mx_room, fbid, fb_receiver, index, fb_chat, timestamp) "
        "SELECT mxid, mx_room, fbid, fb_receiver::bigint, index, fb_chat::bigint, "
        "       (extract(epoch from date) * 1000)::bigint "
        "FROM legacy_message"
    )
    await conn.execute(
        "INSERT INTO reaction (mxid, mx_room, fb_msgid, fb_receiver, fb_sender, reaction) "
        "SELECT mxid, mx_room, fb_msgid, fb_receiver::bigint, fb_sender::bigint, reaction "
        "FROM legacy_reaction"
    )
    try:
        async with conn.transaction():
            await conn.execute(
                'INSERT INTO user_portal ("user", portal, portal_receiver, in_community) '
                'SELECT "user"::bigint, portal::bigint, portal_receiver::bigint, in_community '
                'FROM legacy_user_portal')
            await conn.execute(
                'INSERT INTO user_contact ("user", contact, in_community) '
                'SELECT "user"::bigint, contact::bigint, in_community '
                "FROM legacy_contact")
    except ForeignKeyViolationError:
        pass
    await conn.execute("UPDATE portal SET fb_receiver=0 WHERE fb_type<>'USER'")
    await conn.execute("UPDATE reaction SET fb_receiver=0 WHERE fb_receiver "
                       "IN (SELECT fbid FROM portal WHERE fb_receiver=0)")



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import NamedTuple, Type, Optional, Any
from enum import IntEnum


class TType(IntEnum):
    """
    Thrift type identifiers for the compact struct encoding.

    https://github.com/apache/thrift/blob/master/doc/specs/thrift-compact-protocol.md#struct-encoding
    """
    STOP = 0
    TRUE = 1
    FALSE = 2
    BYTE = 3
    I16 = 4
    I32 = 5
    I64 = 6
    DOUBLE = 7
    BINARY = 8
    LIST = 9
    SET = 10
    MAP = 11
    STRUCT = 12

    # Used internally to represent booleans in schemas.
    BOOL = 0xa1


class RecursiveType(NamedTuple):
    """A type container that can exactly specify the expected types for nested lists/maps."""
    type: TType
    python_type: Optional[Type[Any]] = None
    item_type: Optional['RecursiveType'] = None
    key_type: Optional['RecursiveType'] = None
    value_type: Optional['RecursiveType'] = None



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, List, Dict, Tuple, Union

from yarl import URL
from attr import dataclass
import attr

from mautrix.types import ExtensibleEnum, SerializableAttrs, JSON, Obj, deserializer
from ..common import ThreadFolder, MessageUnsendability


@dataclass
class ParticipantID(SerializableAttrs['ParticipantID']):
    id: str


@dataclass
class ReadReceipt(SerializableAttrs['ReadReceipt']):
    action_timestamp_precise: str
    timestamp_precise: str
    actor: ParticipantID

    @property
    def timestamp(self) -> int:
        return int(self.timestamp_precise)

    @property
    def action_timestamp(self) -> int:
        return int(self.action_timestamp_precise)


@dataclass
class ReadReceiptList(SerializableAttrs['ReadReceiptList']):
    nodes: List[ReadReceipt]


@dataclass
class Picture(SerializableAttrs['Picture']):
    uri: str
    height: Optional[int] = None
    width: Optional[int] = None

    @property
    def dimensions(self) -> Tuple[int, int]:
        return self.width, self.height


class StructuredNamePart(ExtensibleEnum):
    FIRST = "first"
    MIDDLE = "middle"
    LAST = "last"


@dataclass
class StructuredNameChunk(SerializableAttrs['StructuredNameChunk']):
    length: int
    offset: int
    part: StructuredNamePart


@dataclass
class StructuredName(SerializableAttrs['StructuredName']):
    parts: List[StructuredNameChunk]
    phonetic_name: Optional[str]
    text: str

    def to_dict(self) -> Dict[str, str]:
        return {
            f"{part.part.value}_name": self.text[part.offset:part.offset + part.length]
            for part in self.parts
        }


@dataclass
class FriendCount(SerializableAttrs['FriendCount']):
    count: int


@dataclass(kw_only=True)
class MinimalParticipant(ParticipantID, SerializableAttrs['MinimalParticipant']):
    name: Optional[str] = None


class FriendshipStatus(ExtensibleEnum):
    ARE_FRIENDS = "ARE_FRIENDS"
    CAN_REQUEST = "CAN_REQUEST"
    CANNOT_REQUEST = "CANNOT_REQUEST"
    INCOMING_REQUEST = "INCOMING_REQUEST"
    OUTGOING_REQUEST = "OUTGOING_REQUEST"


class ReachabilityStatus(ExtensibleEnum):
    REACHABLE = "REACHABLE"
    UNREACHABLE_USER_TYPE = "UNREACHABLE_USER_TYPE"


class ParticipantType(ExtensibleEnum):
    USER = "User"
    PAGE = "Page"


@dataclass(kw_only=True)
class Participant(MinimalParticipant, SerializableAttrs['Participant']):
    typename: ParticipantType = attr.ib(metadata={"json": "__typename"})

    username: Optional[str] = None
    structured_name: Optional[StructuredName] = None
    nickname_for_viewer: Optional[str] = None

    profile_pic_small: Optional[Picture] = None
    profile_pic_medium: Optional[Picture] = None
    profile_pic_large: Optional[Picture] = None

    friends: Optional[FriendCount] = None
    friendship_status: Optional[FriendshipStatus] = None
    mutual_friends: Optional[FriendCount] = None
    reachability_status_type: Optional[ReachabilityStatus] = None
    registration_time: Optional[int] = None

    is_aloha_proxy_confirmed: bool = False
    is_blocked_by_viewer: bool = False
    is_banned_by_page_viewer: bool = False
    is_deactivated_allowed_on_messenger: bool = False
    is_managing_parent_approved_user: bool = False
    is_memorialized: bool = False
    is_message_blocked_by_viewer: bool = False
    is_message_ignored_by_viewer: bool = False
    is_pseudo_blocked_by_viewer: bool = False
    is_messenger_user: bool = False
    is_partial: Optional[bool] = None
    is_verified: bool = False
    is_viewer_friend: bool = False
    can_viewer_message: bool = True


@dataclass
class ParticipantNode(SerializableAttrs['ParticipantNode']):
    id: str
    messaging_actor: Participant


@dataclass
class ParticipantList(SerializableAttrs['ParticipantList']):
    nodes: List[ParticipantNode]


@dataclass
class MessageSender(SerializableAttrs['MessageSender']):
    id: str
    messaging_actor: MinimalParticipant


@dataclass
class MessageRange(SerializableAttrs['MessageRange']):
    length: int
    offset: int
    entity: Optional[ParticipantID] = None

    @property
    def user_id(self) -> str:
        return self.entity.id


class MessagePowerUpType(ExtensibleEnum):
    NONE = "NONE"
    LOVE = "LOVE"
    GIFT_WRAP = "GIFT_WRAP"
    CELEBRATION = "CELEBRATION"
    FIRE = "FIRE"


@dataclass
class MessagePowerUp(SerializableAttrs['MessagePowerUp']):
    style: MessagePowerUpType


@dataclass
class MessageText(SerializableAttrs['MessageText']):
    text: str
    ranges: List[MessageRange] = attr.ib(factory=lambda: [])


@dataclass
class Reaction(SerializableAttrs['Reaction']):
    reaction: str
    reaction_timestamp: int
    user: ParticipantID


@dataclass
class Dimensions(SerializableAttrs['Dimensions']):
    x: int
    y: int


class AttachmentType(ExtensibleEnum):
    IMAGE = "MessageImage"
    ANIMATED_IMAGE = "MessageAnimatedImage"
    FILE = "MessageFile"
    AUDIO = "MessageAudio"
    VIDEO = "MessageVideo"
    LOCATION = "MessageLocation"
    LIVE_LOCATION = "MessageLiveLocation"
    EXTERNAL_URL = "ExternalUrl"
    STORY = "Story"


class ImageType(ExtensibleEnum):
    FILE_ATTACHMENT = "FILE_ATTACHMENT"
    MESSENGER_CAM = "MESSENGER_CAM"
    TRANSPARENT = "TRANSPARENT"


class VideoType(ExtensibleEnum):
    FILE_ATTACHMENT = "FILE_ATTACHMENT"
    RECORDED_VIDEO = "RECORDED_VIDEO"
    SPEAKING_STICKER = "SPEAKING_STICKER"
    RECORDED_STICKER = "RECORDED_STICKER"
    VIDEO_MAIL = "VIDEO_MAIL"
    IG_SELFIE_STICKER = "IG_SELFIE_STICKER"


@dataclass
class Attachment(SerializableAttrs['Attachment']):
    typename: AttachmentType = attr.ib(metadata={"json": "__typename"})
    id: str
    attachment_fbid: str
    filename: str
    mimetype: str
    filesize: Optional[int] = None
    render_as_sticker: bool = False

    image_type: Optional[ImageType] = None
    original_dimensions: Optional[Dimensions] = None
    image_blurred_preview: Optional[Picture] = None
    image_full_screen: Optional[Picture] = None
    image_large_preview: Optional[Picture] = None
    image_medium_preview: Optional[Picture] = None
    image_small_preview: Optional[Picture] = None

    # For animated images
    animated_image_render_as_sticker: bool = False
    animated_image_original_dimensions: Optional[Dimensions] = None
    animated_image_full_screen: Optional[Picture] = None
    animated_image_large_preview: Optional[Picture] = None
    animated_image_medium_preview: Optional[Picture] = None
    animated_image_small_preview: Optional[Picture] = None
    animated_static_image_full_screen: Optional[Picture] = None
    animated_static_image_large_preview: Optional[Picture] = None
    animated_static_image_medium_preview: Optional[Picture] = None
    animated_static_image_small_preview: Optional[Picture] = None

    # For audio files
    is_voicemail: bool = False
    playable_url: Optional[str] = None

    # For audio and video files
    playable_duration_in_ms: Optional[int] = None

    # For video files
    video_type: Optional[VideoType] = None
    streaming_image_thumbnail: Optional[Picture] = attr.ib(default=None, metadata={
        "json": "streamingImageThumbnail"})
    video_filesize: Optional[int] = None
    attachment_video_url: Optional[str] = None


@dataclass
class MinimalSticker(SerializableAttrs['MinimalSticker']):
    # 369239263222822 = "like"
    id: str


@dataclass
class StickerPackMeta(SerializableAttrs['StickerPackMeta']):
    id: str
    is_comments_capable: bool
    is_composer_capable: bool
    is_messenger_capable: bool
    is_messenger_kids_capable: bool
    is_montage_capable: bool
    is_posts_capable: bool
    is_sms_capable: bool


class StickerType(ExtensibleEnum):
    REGULAR = "REGULAR"
    AVATAR = "AVATAR"
    CUSTOM = "CUSTOM"


@dataclass
class Sticker(MinimalSticker, SerializableAttrs['Sticker']):
    pack: StickerPackMeta
    animated_image: Picture
    preview_image: Picture
    thread_image: Picture
    sticker_type: StickerType
    label: Optional[str] = None


@dataclass
class ExtensibleText(SerializableAttrs['ExtensibleText']):
    text: str


@dataclass
class Coordinates(SerializableAttrs['Coordinates']):
    latitude: float
    longitude: float


@dataclass
class StoryTarget(SerializableAttrs['StoryTarget']):
    typename: AttachmentType = attr.ib(metadata={"json": "__typename"})
    id: Optional[str] = None
    url: Optional[str] = None
    coordinates: Optional[Coordinates] = None


@dataclass
class StoryMediaAttachment(SerializableAttrs['StoryMediaAttachment']):
    typename_str: str = attr.ib(metadata={"json": "__typename"})
    id: Optional[str] = None
    width: Optional[int] = None
    height: Optional[int] = None
    owner: Optional[MinimalParticipant] = None
    title: Optional[ExtensibleText] = None
    image: Optional[Picture] = None
    image_natural: Optional[Picture] = attr.ib(default=None, metadata={"json": "imageNatural"})
    image_fullscreen: Optional[Picture] = attr.ib(default=None,
                                                  metadata={"json": "imageFullScreen"})
    image_large: Optional[Picture] = attr.ib(default=None, metadata={"json": "imageLarge"})
    playable_url: Optional[str] = None
    is_looping: bool = False
    is_playable: bool = False


@dataclass
class StoryAttachment(SerializableAttrs['StoryAttachment']):
    title: str
    url: str
    # TODO enum? share, message_location, attached_story, photo, games_app, messenger_native_templates, unavailable, fallback
    style_list: List[str] = attr.ib(factory=lambda: [])
    title_with_entities: Optional[ExtensibleText] = None
    description: Optional[ExtensibleText] = None
    source: Optional[ExtensibleText] = None
    subtitle: Optional[str] = None
    target: Optional[StoryTarget] = None
    deduplication_key: Optional[str] = None
    media: Optional[StoryMediaAttachment] = None

    @property
    def clean_url(self) -> URL:
        url = URL(self.url)
        if url.host == "l.facebook.com":
            url = URL(url.query["u"])
        return url


@dataclass
class ExtensibleAttachment(SerializableAttrs['ExtensibleAttachment']):
    id: str
    is_forwardable: bool
    story_attachment: Optional[StoryAttachment] = None


@dataclass(kw_only=True)
class MinimalMessage(SerializableAttrs['MinimalMessage']):
    # IDs and message are not present in some action messages like adding to group
    id: Optional[str] = None
    message_id: Optional[str] = None
    message: Optional[MessageText] = None
    message_sender: MessageSender
    sticker: Optional[MinimalSticker] = None
    blob_attachments: List[Attachment] = attr.ib(factory=lambda: [])
    extensible_attachment: ExtensibleAttachment = attr.ib(default=None)


class ReplyStatus(ExtensibleEnum):
    VALID = "VALID"
    DELETED = "DELETED"


@dataclass
class Reply(SerializableAttrs['Reply']):
    message: MinimalMessage
    status: ReplyStatus


@dataclass(kw_only=True)
class Message(MinimalMessage, SerializableAttrs['Message']):
    snippet: str
    timestamp_precise: str
    unsent_timestamp_precise: Optional[str] = None
    offline_threading_id: Optional[str] = None
    tags_list: List[str] = attr.ib(factory=lambda: [])
    message_reactions: List[Reaction] = attr.ib(factory=lambda: [])
    replied_to_message: Optional[Reply] = None
    message_unsendability_status: MessageUnsendability = MessageUnsendability.CAN_UNSEND

    is_sponsored: bool = False
    is_user_generated: bool = True
    unread: bool = False
    ttl: Optional[int] = None

    @property
    def timestamp(self) -> int:
        return int(self.timestamp_precise)

    @property
    def unsent_timestamp(self) -> Optional[int]:
        return int(self.unsent_timestamp_precise) if self.unsent_timestamp_precise else None


@dataclass
class PageInfo(SerializableAttrs['PageInfo']):
    has_next_page: bool = False
    has_previous_page: bool = False

    end_cursor: Optional[str] = None
    start_cursor: Optional[str] = None


@dataclass
class MessageList(SerializableAttrs['MessageList']):
    nodes: List[Message]
    # Not present in last_message and other such "lists"
    page_info: PageInfo = attr.ib(factory=lambda: PageInfo())


@dataclass(eq=True, frozen=True)
class ThreadKey(SerializableAttrs['ThreadKey']):
    other_user_id: Optional[str] = None
    thread_fbid: Optional[str] = None

    @property
    def id(self) -> Optional[int]:
        if self.other_user_id:
            return int(self.other_user_id)
        elif self.thread_fbid:
            return int(self.thread_fbid)
        else:
            return None


@dataclass(kw_only=True)
class Thread(SerializableAttrs['Thread']):
    id: str
    folder: ThreadFolder
    name: Optional[str]
    thread_key: ThreadKey
    image: Optional[Picture]

    messages_count: int
    unread_count: int
    unsend_limit: int
    privacy_mode: int
    thread_pin_timestamp: int
    thread_queue_enabled: bool
    thread_unsendability_status: MessageUnsendability
    update_time_precise: Optional[str] = None

    last_message: MessageList
    messages: MessageList
    read_receipts: ReadReceiptList
    all_participants: ParticipantList

    thread_admins: List[ParticipantID]

    is_admin_supported: bool
    is_business_page_active: bool
    is_disappearing_mode: bool
    is_fuss_red_page: bool
    is_group_thread: bool
    is_ignored_by_viewer: bool
    is_pinned: bool
    is_viewer_allowed_to_add_members: bool
    is_viewer_subscribed: bool
    can_viewer_reply: bool
    can_participants_claim_admin: bool


@dataclass
class ThreadListResponse(SerializableAttrs['ThreadListResponse']):
    count: int
    unread_count: int
    unseen_count: int
    mute_until: int
    nodes: List[Thread]
    page_info: PageInfo
    sync_sequence_id: str


@dataclass
class ThreadQueryResponse(SerializableAttrs['ThreadQueryResponse']):
    message_threads: List[Thread]


@dataclass
class StickerPreviewResponse(SerializableAttrs['StickerPreviewResponse']):
    nodes: List[Sticker]


@dataclass
class MessageUnsendResponse(SerializableAttrs['MessageUnsendResponse']):
    did_succeed: bool
    error_code: str
    error_message: str


@dataclass
class ImageFragment(SerializableAttrs['ImageFragment']):
    typename: AttachmentType = attr.ib(metadata={"json": "__typename"})
    id: str
    animated_gif: Optional[Picture] = None
    image: Optional[Picture] = None


@dataclass
class SubsequentMediaNode(SerializableAttrs['SubsequentMediaNode']):
    typename: AttachmentType = attr.ib(metadata={"json": "__typename"})
    id: str
    legacy_attachment_id: str
    creation_time: int
    creator: MinimalParticipant
    adjusted_size: Optional[Picture] = None
    image_thumbnail: Optional[Picture] = attr.ib(default=None, metadata={"json": "imageThumbnail"})
    media_url: Optional[str] = attr.ib(default=None, metadata={"json": "mediaUrl"})
    original_dimensions: Optional[Dimensions] = None


@dataclass
class SubsequentMediaResponse(SerializableAttrs['SubsequentMediaResponse']):
    nodes: List[SubsequentMediaNode]
    page_info: PageInfo


@dataclass
class FileAttachmentWithURL(SerializableAttrs['FileAttachmentWithURL']):
    typename: AttachmentType = attr.ib(metadata={"json": "__typename"})
    attachment_fbid: str
    id: str
    url: str


@dataclass
class FileAttachmentURLResponse(SerializableAttrs['FileAttachmentURLResponse']):
    id: str
    blob_attachments: List[FileAttachmentWithURL]


@dataclass(kw_only=True)
class OwnInfo(SerializableAttrs['OwnInfo']):
    id: str
    birthday: Optional[str] = None
    gender: Optional[str] = None
    locale: Optional[str] = None
    email: Optional[str] = None
    name: str
    first_name: Optional[str] = None
    middle_name: Optional[str] = None
    last_name: Optional[str] = None
    link: Optional[str] = None
    is_employee: bool = False
    verified: bool = False
    published_timeline: bool = False
    timezone: int = 0
    updated_time: Optional[str] = None


@dataclass
class MessageSearchResult(SerializableAttrs['MessageSearchResult']):
    thread_id: str
    name: Optional[str]
    # TODO message_thread, matched_message


# TODO there might be other types of search results
SearchResultNode = Union[MessageSearchResult, Participant, Obj]


@deserializer(SearchResultNode)
def deserialize_search_node(val: JSON) -> SearchResultNode:
    type = val["__typename"]
    if type in ("User", "Page"):
        return Participant.deserialize(val)
    elif type == "MessageSearchResult":
        return MessageSearchResult.deserialize(val)
    return Obj(**val)


@dataclass
class SearchResult(SerializableAttrs['SearchResult']):
    node: SearchResultNode
    # there's a header field with some random metadata too


@dataclass
class SearchResults(SerializableAttrs['SearchResults']):
    edges: List[SearchResult]


@dataclass
class SearchEntitiesResponse(SerializableAttrs['SearchEntitiesResponse']):
    cache_id: str
    search_results: SearchResults



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any, Union, List, Collection, Dict, Optional, TYPE_CHECKING
import struct
import io

from .type import TType

if TYPE_CHECKING:
    from .autospec import ThriftObject


class ThriftWriter(io.BytesIO):
    """
    ThriftWriter implements encoding Python values into the Thrift Compact protocol.

    https://github.com/apache/thrift/blob/master/doc/specs/thrift-compact-protocol.md
    """
    _prev_field_id: int
    _stack: List[int]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._prev_field_id = 0
        self._stack = []

    def _push_stack(self) -> None:
        self._stack.append(self._prev_field_id)
        self._prev_field_id = 0

    def _pop_stack(self) -> None:
        if self._stack:
            self._prev_field_id = self._stack.pop()

    def _write_byte(self, byte: Union[int, TType]) -> None:
        self.write(bytes([byte]))

    @staticmethod
    def _to_zigzag(val: int, bits: int) -> int:
        return (val << 1) ^ (val >> (bits - 1))

    def _write_varint(self, val: int) -> None:
        while True:
            byte = val & ~0x7f
            if byte == 0:
                self._write_byte(val)
                break
            elif byte == -128:
                self._write_byte(0)
                break
            else:
                self._write_byte((val & 0xff) | 0x80)
                val = val >> 7

    def _write_word(self, val: int) -> None:
        self._write_varint(self._to_zigzag(val, 16))

    def _write_int(self, val: int) -> None:
        self._write_varint(self._to_zigzag(val, 32))

    def _write_long(self, val: int) -> None:
        self._write_varint(self._to_zigzag(val, 64))

    def _write_field_begin(self, field_id: int, ttype: TType) -> None:
        ttype_val = ttype.value
        delta = field_id - self._prev_field_id
        if 0 < delta < 16:
            self._write_byte((delta << 4) | ttype_val)
        else:
            self._write_byte(ttype_val)
            self._write_word(field_id)
        self._prev_field_id = field_id

    def _write_string(self, val: Union[str, bytes]) -> None:
        if isinstance(val, str):
            val = val.encode("utf-8")
        self._write_varint(len(val))
        self.write(val)

    def write_map(self, field_id: int, key_type: TType, value_type: TType, val: Dict[Any, Any]
                  ) -> None:
        self._write_field_begin(field_id, TType.MAP)
        if not map:
            self._write_byte(0)
            return
        self._write_varint(len(val))
        self._write_byte(((key_type.value & 0xf) << 4) | (value_type.value & 0xf))
        for key, value in val.items():
            self.write_val(None, key_type, key)
            self.write_val(None, value_type, value)

    def write_stop(self) -> None:
        self._write_byte(TType.STOP.value)
        self._pop_stack()

    def write_list(self, field_id: int, item_type: TType, val: Collection[Any]) -> None:
        self._write_field_begin(field_id, TType.LIST)
        if len(val) < 0x0f:
            self._write_byte((len(val) << 4) | item_type.value)
        else:
            self._write_byte(0xf0 | item_type.value)
            self._write_varint(len(val))
        for item in val:
            self.write_val(None, item_type, item)

    def write_struct_begin(self, field_id: int) -> None:
        self._write_field_begin(field_id, TType.STRUCT)
        self._push_stack()

    def write_val(self, field_id: Optional[int], ttype: TType, val: Any) -> None:
        if ttype == TType.BOOL:
            if field_id is None:
                raise ValueError("booleans can only be used in structs")
            self._write_field_begin(field_id, TType.TRUE if val else TType.FALSE)
            return
        if field_id is not None:
            self._write_field_begin(field_id, ttype)
        if ttype == TType.BYTE:
            self._write_byte(val)
        elif ttype == TType.I16:
            self._write_word(val)
        elif ttype == TType.I32:
            self._write_int(val)
        elif ttype == TType.I64:
            self._write_long(val)
        elif ttype == TType.DOUBLE:
            self.write(struct.pack("<d", val))
        elif ttype == TType.BINARY:
            self._write_string(val)
        else:
            raise ValueError(f"{ttype} is not supported by write_val()")

    def write_struct(self, obj: 'ThriftObject') -> None:
        for field_id in iter(obj.thrift_spec):
            meta = obj.thrift_spec[field_id]

            val = getattr(obj, meta.name, None)
            if val is None:
                continue

            rtype = meta.rtype
            if rtype.type in (TType.LIST, TType.SET):
                self.write_list(field_id, rtype.item_type.type, val)
            elif rtype.type == TType.MAP:
                self.write_map(field_id, rtype.key_type.type, rtype.value_type.type, val)
            elif rtype.type == TType.STRUCT:
                self.write_struct_begin(field_id)
                self.write_struct(val)
            else:
                self.write_val(field_id, rtype.type, val)
        self.write_stop()



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, Union
from enum import Enum
import pkgutil
import json

_raw_topic_map: Dict[str, int] = json.loads(pkgutil.get_data("maufbapi.mqtt", "topics.json"))
# Mapping from name to numeric ID
topic_map: Dict[str, str] = {key: str(value) for key, value in _raw_topic_map.items()}
# Mapping from numeric ID to name
_reverse_topic_map: Dict[str, str] = {value: key for key, value in topic_map.items()}


class RealtimeTopic(Enum):
    SYNC_CREATE_QUEUE = "/messenger_sync_create_queue"
    MESSAGE_SYNC = "/t_ms"
    SEND_MESSAGE = "/t_sm"
    SEND_MESSAGE_RESP = "/t_sm_rp"
    REGION_HINT = "/t_region_hint"
    MARK_THREAD_READ = "/t_mt_req"
    MARK_THREAD_READ_RESPONSE = "/t_mt_resp"
    OPENED_THREAD = "/opened_thread"

    @property
    def encoded(self) -> str:
        return topic_map[self.value]

    @staticmethod
    def decode(val: str) -> Union['RealtimeTopic', str]:
        topic = _reverse_topic_map[val]
        try:
            return RealtimeTopic(topic)
        except ValueError:
            return topic
